# -*- coding: utf-8 -*-
import datetime

import numpy as np

from ..probability import PDAHypothesiser
from ...types.detection import Detection, MissedDetection
from ...types.numeric import Probability
from ...types.state import GaussianState
from ...types.track import Track


def test_pda(predictor, updater):

    timestamp = datetime.datetime.now()
    track = Track([GaussianState(np.array([[0]]), np.array([[1]]), timestamp)])
    detection1 = Detection(np.array([[2]]))
    detection2 = Detection(np.array([[8]]))
    detections = {detection1, detection2}

    hypothesiser = PDAHypothesiser(predictor, updater,
                                   clutter_spatial_density=1.2e-2,
                                   prob_detect=0.9, prob_gate=0.99)

    mulltihypothesis = \
        hypothesiser.hypothesise(track, detections, timestamp)

    # There are 3 weighted hypotheses - Detections 1 and 2, MissedDectection
    assert len(mulltihypothesis) == 3

    # Each hypothesis has a probability/weight attribute
    assert all(hypothesis.probability >= 0 and
               isinstance(hypothesis.probability, Probability)
               for hypothesis in
               mulltihypothesis)

    #  Detection 1, 2 and MissedDetection are present
    assert any(hypothesis.measurement is detection1 for hypothesis in
               mulltihypothesis)
    assert any(hypothesis.measurement is detection2 for hypothesis in
               mulltihypothesis)
    assert any(isinstance(hypothesis.measurement, MissedDetection)
               for hypothesis in
               mulltihypothesis)



# -*- coding: utf-8 -*-
from .base import Hypothesiser
from ..base import Property
from ..measures import Measure
from ..predictor import Predictor
from ..types.multihypothesis import \
    MultipleHypothesis
from ..types.hypothesis import SingleDistanceHypothesis
from ..types.detection import MissedDetection
from ..updater import Updater


class DistanceHypothesiser(Hypothesiser):
    """Prediction Hypothesiser based on a Measure

    Generate track predictions at detection times and score each hypothesised
    prediction-detection pair using the distance of the supplied
    :class:`Measure` class.
    """

    predictor = Property(
        Predictor,
        doc="Predict tracks to detection times")
    updater = Property(
        Updater,
        doc="Updater used to get measurement prediction")
    measure = Property(
        Measure,
        doc="Measure class used to calculate the distance between two states.")
    missed_distance = Property(
        float,
        default=float('inf'),
        doc="Distance for a missed detection. Default is set to infinity")

    def hypothesise(self, track, detections, timestamp):
        """ Evaluate and return all track association hypotheses.

        For a given track and a set of N available detections, return a
        MultipleHypothesis object with N+1 detections (first detection is
        a 'MissedDetection'), each with an associated distance measure..

        Parameters
        ----------
        track: :class:`~.Track`
            The track object to hypothesise on
        detections: :class:`list`
            A list of :class:`~Detection` objects, representing the available
            detections.
        timestamp: :class:`datetime.datetime`
            A timestamp used when evaluating the state and measurement
            predictions. Note that if a given detection has a non empty
            timestamp, then prediction will be performed according to
            the timestamp of the detection.

        Returns
        -------
        : :class:`~.MultipleHypothesis`
            A container of :class:`~SingleDistanceHypothesis` objects

        """
        hypotheses = list()

        # Common state & measurement prediction
        prediction = self.predictor.predict(track.state, timestamp=timestamp)
        measurement_prediction = self.updater.get_measurement_prediction(
            prediction)

        # Missed detection hypothesis with distance as 'missed_distance'
        hypotheses.append(
            SingleDistanceHypothesis(
                prediction,
                MissedDetection(timestamp=timestamp),
                self.missed_distance,
                measurement_prediction))

        # True detection hypotheses
        for detection in detections:

            # Re-evaluate prediction
            prediction = self.predictor.predict(
                track.state, timestamp=detection.timestamp)

            # Compute measurement prediction and distance measure
            measurement_prediction = self.updater.get_measurement_prediction(
                prediction, detection.measurement_model)
            distance = self.measure(measurement_prediction, detection)

            # True detection hypothesis
            hypotheses.append(
                SingleDistanceHypothesis(
                    prediction,
                    detection,
                    distance,
                    measurement_prediction))

        return MultipleHypothesis(sorted(hypotheses, reverse=True))



import numpy as np
from scipy.stats import multivariate_normal

from .base import Initiator, GaussianInitiator
from ..base import Property
from ..models.measurement import MeasurementModel
from ..types.hypothesis import SingleHypothesis
from ..types.numeric import Probability
from ..types.particle import Particle
from ..types.state import GaussianState
from ..types.track import Track
from ..types.update import GaussianStateUpdate, ParticleStateUpdate
from ..updater.kalman import KalmanUpdater


class SinglePointInitiator(GaussianInitiator):
    """ SinglePointInitiator class"""

    prior_state = Property(GaussianState, doc="Prior state information")
    measurement_model = Property(MeasurementModel, doc="Measurement model")

    def initiate(self, unassociated_detections, **kwargs):
        """Initiates tracks given unassociated measurements

        Parameters
        ----------
        unassociated_detections : list of \
        :class:`stonesoup.types.detection.Detection`
            A list of unassociated detections

        Returns
        -------
        : :class:`sets.Set` of :class:`stonesoup.types.track.Track`
            A list of new tracks with an initial :class:`~.GaussianState`
        """

        updater = KalmanUpdater(self.measurement_model)

        tracks = set()
        for detection in unassociated_detections:
            measurement_prediction = updater.get_measurement_prediction(
                self.prior_state, detection.measurement_model)
            track_state = updater.update(SingleHypothesis(
                self.prior_state, detection, measurement_prediction))
            track = Track([track_state])
            tracks.add(track)

        return tracks


class LinearMeasurementInitiator(GaussianInitiator):
    """Initiator that maps measurement space to state space

    This initiator utilises the :class:`~.MeasurementModel` matrix to convert
    :class:`~.Detection` state vector and model covariance into state space.
    This then replaces mapped values in the :attr:`prior_state` to form the
    initial :class:`~.GaussianState` of the :class:`~.Track`.
    """
    prior_state = Property(GaussianState, doc="Prior state information")
    measurement_model = Property(MeasurementModel, doc="Measurement model")

    def initiate(self, detections, **kwargs):
        tracks = set()

        model_matrix = self.measurement_model.matrix()
        model_covar = self.measurement_model.covar()

        prior_state_vector = self.prior_state.state_vector.copy()
        prior_covar = self.prior_state.covar.copy()

        # Zero out elements of prior state that will be replaced by measurement
        mapped_dimensions, _ = np.nonzero(
            model_matrix.T@np.ones((model_matrix.shape[0], 1)))
        prior_state_vector[mapped_dimensions, :] = 0
        prior_covar[mapped_dimensions, :] = 0

        inv_model_matrix = np.linalg.pinv(model_matrix)

        for detection in detections:
            tracks.add(Track([GaussianStateUpdate(
                prior_state_vector + inv_model_matrix@detection.state_vector,
                prior_covar
                + inv_model_matrix@model_covar@model_matrix.astype(bool),
                SingleHypothesis(None, detection),
                timestamp=detection.timestamp)
            ]))
        return tracks


class GaussianParticleInitiator(Initiator):
    """Gaussian Particle Initiator class

    Utilising Gaussian Initiator, sample from the resultant track's state
    to generate a number of particles, overwriting with a
    :class:`~.ParticleState`.
    """

    initiator = Property(
        GaussianInitiator,
        doc="Gaussian Initiator which will be used to generate tracks.")
    number_particles = Property(
        float, default=200, doc="Number of particles for initial track state")

    def initiate(self, unassociated_detections, **kwargs):
        """Initiates tracks given unassociated measurements

        Parameters
        ----------
        unassociated_detections : list of :class:`~.Detection`
            A list of unassociated detections

        Returns
        -------
        : set of :class:`~.Track`
            A list of new tracks with a initial :class:`~.ParticleState`
        """
        tracks = self.initiator.initiate(unassociated_detections, **kwargs)
        weight = Probability(1/self.number_particles)
        for track in tracks:
            samples = multivariate_normal.rvs(track.state_vector.ravel(),
                                              track.covar,
                                              size=self.number_particles)
            particles = [
                Particle(sample.reshape(-1, 1), weight=weight)
                for sample in samples]
            track[-1] = ParticleStateUpdate(
                particles,
                track.hypothesis,
                timestamp=track.timestamp)

        return tracks



# coding: utf-8
import datetime

import scipy as sp
from scipy.stats import multivariate_normal

from ..linear import SingerApproximate, CombinedLinearGaussianTransitionModel


def test_singer1dmodel_approximate():
    """ SingerApproximate 1D Transition Model test for small timediff """
    state_vec = sp.array([[3.0], [1.0], [0.1]])
    noise_diff_coeffs = sp.array([0.01])
    recips_decorr_times = sp.array([0.1])
    base(state_vec, noise_diff_coeffs, recips_decorr_times, timediff=0.4)


def test_singer2dmodel_approximate():
    """ SingerApproximate 2D Transition Model test for small timediff """
    state_vec = sp.array([[3.0], [1.0], [0.1],
                          [2.0], [2.0], [0.2]])
    noise_diff_coeffs = sp.array([0.01, 0.02])
    recips_decorr_times = sp.array([0.1, 0.1])

    base(state_vec, noise_diff_coeffs, recips_decorr_times, timediff=0.4)


def test_singer3dmodel_approximate():
    """ SingerApproximate 3D Transition Model test for small timediff """
    state_vec = sp.array([[3.0], [1.0], [0.1],
                          [2.0], [2.0], [0.2],
                          [4.0], [0.5], [0.05]])
    noise_diff_coeffs = sp.array([0.01, 0.02, 0.005])
    recips_decorr_times = sp.array([0.1, 0.1, 0.1])
    base(state_vec, noise_diff_coeffs, recips_decorr_times, timediff=0.4)


def base(state_vec, noise_diff_coeffs, recips_decorr_times, timediff=1.0):
    """ Base test for n-dimensional ConstantAcceleration Transition Models """

    # Create a 1D Singer or an n-dimensional
    # CombinedLinearGaussianTransitionModel object
    dim = len(state_vec) // 3  # pos, vel, acc for each dimension
    if dim == 1:
        model_obj = SingerApproximate(noise_diff_coeff=noise_diff_coeffs[0],
                                      recip_decorr_time=recips_decorr_times[0])
    else:
        model_list = [
            SingerApproximate(
                noise_diff_coeff=noise_diff_coeffs[i],
                recip_decorr_time=recips_decorr_times[i])
            for i in range(0, dim)
        ]
        model_obj = CombinedLinearGaussianTransitionModel(model_list)

    # State related variables
    state_vec = state_vec
    old_timestamp = datetime.datetime.now()
    new_timestamp = old_timestamp + datetime.timedelta(seconds=timediff)
    time_interval = new_timestamp - old_timestamp

    # Model-related components
    noise_diff_coeffs = noise_diff_coeffs  # m/s^3
    mat_list = []
    covar_list = []

    for i in range(0, dim):
        recip_decorr_time = recips_decorr_times[i]
        recip_decorr_timedt = recip_decorr_time * timediff
        noise_diff_coeff = noise_diff_coeffs[i]

        mat_list.append(sp.array(
            [[1,
              timediff,
              (recip_decorr_timedt - 1 + sp.exp(-recip_decorr_timedt)) /
              sp.power(recip_decorr_time, 2)],
             [0,
              1,
              (1 - sp.exp(-recip_decorr_timedt)) / recip_decorr_time],
             [0,
              0,
              sp.exp(-recip_decorr_timedt)]]))

        covar_list.append(sp.array(
            [[sp.power(timediff, 5) / 20,
              sp.power(timediff, 4) / 8,
              sp.power(timediff, 3) / 6],
             [sp.power(timediff, 4) / 8,
              sp.power(timediff, 3) / 3,
              sp.power(timediff, 2) / 2],
             [sp.power(timediff, 3) / 6,
              sp.power(timediff, 2) / 2,
              timediff]]) * noise_diff_coeff)

    F = sp.linalg.block_diag(*mat_list)
    Q = sp.linalg.block_diag(*covar_list)

    # Ensure ```model_obj.transfer_function(time_interval)``` returns F
    assert sp.array_equal(F, model_obj.matrix(
        timestamp=new_timestamp, time_interval=time_interval))

    # Ensure ```model_obj.covar(time_interval)``` returns Q
    assert sp.array_equal(Q, model_obj.covar(
        timestamp=new_timestamp, time_interval=time_interval))

    # Propagate a state vector throught the model
    # (without noise)
    new_state_vec_wo_noise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=0)
    assert sp.array_equal(new_state_vec_wo_noise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (without noise)
    prob = model_obj.pdf(new_state_vec_wo_noise,
                         state_vec,
                         timestamp=new_timestamp,
                         time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_wo_noise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throughout the model
    # (with internal noise)
    new_state_vec_w_inoise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval)
    assert not sp.array_equal(new_state_vec_w_inoise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model_obj.pdf(new_state_vec_w_inoise,
                         state_vec,
                         timestamp=new_timestamp,
                         time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_inoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throught the model
    # (with external noise)
    noise = model_obj.rvs(timestamp=new_timestamp, time_interval=time_interval)
    new_state_vec_w_enoise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=noise)
    assert sp.array_equal(new_state_vec_w_enoise, F@state_vec+noise)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model_obj.pdf(new_state_vec_w_enoise, state_vec,
                         timestamp=new_timestamp, time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_enoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)



# -*- coding: utf-8 -*-
import datetime

import pytest

from ...types.detection import Detection


@pytest.fixture()
def detector():
    class Detector:
        def detections_gen(self):
            time = datetime.datetime(2019, 4, 1, 14)
            time_step = datetime.timedelta(seconds=1)

            yield time, {
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'green'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'blue'}),
            }

            time += time_step
            yield time, {
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'green'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'green'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'blue'}),
            }

            time += time_step
            yield time, {
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time - time_step/2,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'blue'}),
            }

            time += time_step
            yield time, {
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'blue'}),
            }

            time += time_step
            yield time, {
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'blue'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={}),
            }

            time += time_step*2
            yield time, {
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'blue'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={}),
                Detection([[0], [0]], timestamp=time,
                          metadata={}),
            }

            time -= time_step*1
            yield time, {
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'red'}),
                Detection([[0], [0]], timestamp=time,
                          metadata={'colour': 'blue'}),
            }

    return Detector()



# -*- coding: utf-8 -*-
from pathlib import Path

from ..base import Property
from ..serialise import YAML
from .base import DetectionReader, GroundTruthReader, SensorDataReader


class YAMLReader(DetectionReader, GroundTruthReader, SensorDataReader):
    """YAML Detection Writer"""
    path = Property(Path, doc="File to read data from")

    def __init__(self, path, *args, **kwargs):
        if not isinstance(path, Path):
            path = Path(path)  # Ensure Path
        super().__init__(path, *args, **kwargs)
        self._yaml = YAML()
        self._detections = set()
        self._groundtruth_paths = set()
        self._sensor_data = set()
        self._tracks = set()

    @property
    def detections(self):
        return self._detections

    @property
    def groundtruth_paths(self):
        return self._groundtruth_paths

    @property
    def sensor_data(self):
        return self._sensor_data

    @property
    def tracks(self):
        return self._tracks

    def detections_gen(self):
        for time, _ in self.data_gen():
            yield time, self.detections

    def groundtruth_paths_gen(self):
        for time, _ in self.data_gen():
            yield time, self.groundtruth_paths

    def sensor_data_gen(self):
        for time, _ in self.data_gen():
            yield time, self.sensor_data

    def tracks_gen(self):
        for time, _ in self.data_gen():
            yield time, self.tracks

    def data_gen(self):
        for document in self._yaml.load_all(self.path):
            self._detections = document.get('detections', set())
            self._groundtruth_paths = document.get('groundtruth_paths', set())
            self._sensor_data = document.get('sensor_data', set())
            self._tracks = document.get('tracks', set())
            yield document.pop('time'), document



# -*- coding: utf-8 -*-
"""Provides base for Stone Soup components.

To aid creation of components in Stone Soup, a declarative approach is used to
declare properties of components. These declared properties are then used to
generate the signature for the class, populate documentation, and generate
forms for the user interface.

An example would be:

.. code-block:: python

    class Foo(Base):
        '''Example Foo class'''
        foo = Property(str, doc="foo string parameter")
        bar = Property(int, default=10, doc="bar int parameter, default is 10")


This is equivalent to the following:

.. code-block:: python

    class Foo:
        '''Example Foo class

        Parameters
        ----------
        foo : str
            foo string parameter
        bar : int, optional
            bar int parameter, default is 10
        '''

        def __init__(self, foo, bar=10):
            self.foo = foo
            self.bar = 10

.. note::

    The init method is actually part of :class:`Base` class so in the case of
    having to customise initialisation, :func:`super` should be used e.g.:

    .. code-block:: python

        class Foo(Base):
        '''Example Foo class'''
        foo = Property(str, doc="foo string parameter")
        bar = Property(int, default=10, doc="bar int parameter, default is 10")

        def __init__(self, foo, bar=bar.default, *args, **kwargs):
            if bar < 0:
                raise ValueError("...")
            super().__init__(foo, bar, *args, **kwargs)


"""
import inspect
import sys
from abc import ABCMeta
from collections import OrderedDict
from types import MappingProxyType


class Property:
    """Property(cls, default=inspect.Parameter.empty)
    Property class for definition of attributes on component classes.

    A class must be provided such that the framework is aware of how components
    are put together to create a valid run within the framework. Additionally,
    the class is used by the user interface to generate configuration options
    to the users. The class is not used for any type checking, in the spirit of
    Python's duck typing.

    A default value can be specified to signify the property on the class is
    optional. As ``None`` and ``False`` are reasonable default values,
    :class:`inspect.Parameter.empty` is used to signify the argument is
    mandatory. (Also aliased to :attr:`Property.empty` for ease)

    A description string can also be provided which will be rendered in the
    documentation.

    Parameters
    ----------
    cls : class
        A Python class.
    default : any, optional
        A default value, which should be same type as class or None. Defaults
        to :class:`inspect.Parameter.empty` (alias :attr:`Property.empty`)
    doc : str, optional
        Doc string for property

    Attributes
    ----------
    cls
    default
    doc
    empty : :class:`inspect.Parameter.empty`
        Alias to :class:`inspect.Parameter.empty`
    """
    empty = inspect.Parameter.empty
    _property_name = None

    def __init__(self, cls, *, default=inspect.Parameter.empty, doc=None):
        self.cls = cls
        self.default = default
        self.doc = doc

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return getattr(instance, self._property_name)

    def __set__(self, instance, value):
        setattr(instance, self._property_name, value)

    def __delete__(self, instance):
        delattr(instance, self._property_name)

    def __set_name__(self, owner, name):
        if not isinstance(owner, BaseMeta):
            raise AttributeError("Cannot use Property on this class type")
        self._property_name = "_property_{}".format(name)


class BaseMeta(ABCMeta):
    """Base metaclass for Stone Soup components.

    This metaclass enables the use of the :class:`Property` class to define
    attributes of a class. This includes generation of the init method
    signature.

    The init method signature if defined on a class, the arguments must match
    as declared. However, keyword only arguments can be added to the init
    method if required, as these won't effect the use of the class in the
    framework.
    """

    @classmethod
    def __prepare__(mcls, name, bases, **kwargs):
        return OrderedDict()

    def __new__(mcls, name, bases, namespace):
        if '__init__' not in namespace:
            # Must replace init so we don't overwrite parent class's
            # and blank line below so this doesn't become its docstring!

            def __init__(self, *args, **kwargs):
                super(cls, self).__init__(*args, **kwargs)
            namespace['__init__'] = __init__
        cls = super().__new__(mcls, name, bases, namespace)

        cls._subclasses = set()
        cls._properties = OrderedDict()
        # Update subclass lists, and update properties (in reverse order)
        for bcls in reversed(cls.mro()[1:]):
            if type(bcls) is mcls:
                bcls._subclasses.add(cls)
                cls._properties.update(bcls._properties)
        cls._properties.update(
            (key, value) for key, value in namespace.items()
            if isinstance(value, Property))
        # Optional arguments must follow mandatory
        for name in list(cls._properties):
            if cls._properties[name].default is not Property.empty:
                cls._properties.move_to_end(name)

        if sys.version_info <= (3, 6):  # pragma: no cover
            for name, property_ in cls._properties.items():
                property_.__set_name__(cls, name)

        cls._validate_init()
        cls._generate_signature()

        return cls

    def _validate_init(cls):
        """Validates custom init's arguments."""
        init_signature = inspect.signature(cls.__init__)

        declared_names = list(cls._properties)
        positional_names = [
            parameter.name
            for parameter in init_signature.parameters.values()
            if parameter.kind in (
                parameter.POSITIONAL_ONLY,
                parameter.POSITIONAL_OR_KEYWORD)][1:]  # Ignore 'self' (item 0)
        if positional_names != declared_names[:len(positional_names)]:
            raise TypeError("init arguments don't match declared properties: "
                            "arguments do not match or wrong order")

        has_var_positional = any(
            parameter.kind == parameter.VAR_POSITIONAL
            for parameter in init_signature.parameters.values())
        has_var_keyword = any(
            parameter.kind == parameter.VAR_KEYWORD
            for parameter in init_signature.parameters.values())
        if positional_names != declared_names and not (
                    has_var_positional and has_var_keyword):
            raise TypeError("init arguments don't match declared properties: "
                            "missing argument (or *args and **kwargs missing)")

        keyword_parameters = [
            parameter
            for parameter in init_signature.parameters.values()
            if parameter.kind == parameter.KEYWORD_ONLY]
        if any(parameter.default is parameter.empty
               for parameter in keyword_parameters):
            raise TypeError("new keyword arguments must have default value")

    def _generate_signature(cls):
        """Generates __init__ signature with declared properties."""
        init_signature = inspect.signature(cls.__init__)
        parameters = [next(iter(init_signature.parameters.values()))]  # 'self'
        parameters.extend(
            inspect.Parameter(
                name, inspect.Parameter.POSITIONAL_OR_KEYWORD,
                default=property_.default)
            for name, property_ in cls._properties.items())
        parameters.extend(
            parameter
            for parameter in init_signature.parameters.values()
            if parameter.kind == parameter.KEYWORD_ONLY)
        cls.__init__.__signature__ = init_signature.replace(
            parameters=parameters)

    @property
    def subclasses(cls):
        """Set of subclasses for the class"""
        return frozenset(cls._subclasses)

    @property
    def properties(cls):
        """Set of properties required to initialise the class"""
        return MappingProxyType(cls._properties)


class Base(metaclass=BaseMeta):
    """Base class for framework components.

    This is the base class which should be used for any Stone Soup components.
    Building on the :class:`BaseMeta` this provides a init method which
    populates the declared properties with their values.

    Subclasses can override this method, but they should either call this via
    :func:`super()` or ensure they manually populated the properties as
    declared."""

    def __init__(self, *args, **kwargs):
        init_signature = inspect.signature(self.__init__)
        bound_arguments = init_signature.bind(*args, **kwargs)
        bound_arguments.apply_defaults()
        for name, value in bound_arguments.arguments.items():
            setattr(self, name, value)

    def __repr__(self):
        params = ("{}={!r}".format(name, getattr(self, name))
                  for name in type(self).properties)
        return "{}({})".format(type(self).__name__, ", ".join(params))



# coding: utf-8
import numpy as np

from ..linear import LinearGaussianTimeInvariantTransitionModel


def test_linear_gaussian():
    F = np.eye(3)
    Q = np.eye(3)
    model = LinearGaussianTimeInvariantTransitionModel(
        transition_matrix=F, covariance_matrix=Q)

    x_1 = np.ones([3, 1])
    x_2 = F @ x_1

    assert F.shape[0] == model.ndim_state
    assert np.array_equal(F, model.matrix())
    assert np.array_equal(Q, model.covar())
    assert np.array_equal(x_2, model.function(x_1, noise=np.zeros([3, 1])))
    assert isinstance(model.rvs(), np.ndarray)
    assert isinstance(model.pdf(x_2, x_1), np.float64)



# -*- coding: utf-8 -*-
import datetime

import numpy as np
import pytest

from ...types.array import StateVector
from ...types.detection import MissedDetection, GaussianDetection
from ...types.hypothesis import SingleDistanceHypothesis, \
    SingleProbabilityHypothesis
from ...types.multihypothesis import MultipleHypothesis
from ...types.prediction import StateMeasurementPrediction, \
    GaussianStatePrediction
from ...types.track import Track
from ...types.update import GaussianStateUpdate


@pytest.fixture()
def initiator():
    class TestInitiator:
        def initiate(self, unassociated_detections):
            return {Track([detection])
                    for detection in unassociated_detections}
    return TestInitiator()


@pytest.fixture()
def deleter():
    class TestDeleter:
        def delete_tracks(self, tracks):
            return {track
                    for track in tracks
                    if len(track.states) > 10}
    return TestDeleter()


@pytest.fixture()
def detector():
    class TestDetector:
        def detections_gen(self):
            time = datetime.datetime(2018, 1, 1, 14)
            for step in range(20):
                yield time, {GaussianDetection(
                    StateVector([[step + 10*i]]), [[10]], timestamp=time)
                             for i in range(3)
                             if (step - i) % 5}
                time += datetime.timedelta(minutes=1)
            for step in range(3):
                yield time, set()
                time += datetime.timedelta(minutes=1)
    return TestDetector()


@pytest.fixture()
def data_associator():
    class TestDataAssociator:
        def associate(self, tracks, detections, time):
            associations = {}
            for track in tracks:
                prediction = GaussianStatePrediction(track.state_vector + 1,
                                                     [[5]], time)
                measurement_prediction = StateMeasurementPrediction(
                    prediction.state_vector)
                for detection in detections:
                    if np.array_equal(measurement_prediction.state_vector,
                                      detection.state_vector):
                        associations[track] = \
                            SingleDistanceHypothesis(
                            prediction, detection, 0, measurement_prediction)
                        break
                else:
                    associations[track] = SingleDistanceHypothesis(
                        prediction, None, None, 1)
            return associations
    return TestDataAssociator()


@pytest.fixture()
def data_mixture_associator():
    class TestDataMixtureAssociator:
        def associate(self, tracks, detections, time):
            associations = {}
            for track in tracks:
                prediction = GaussianStatePrediction(track.state_vector + 1,
                                                     [[5]], time)
                measurement_prediction = StateMeasurementPrediction(
                    prediction.state_vector)
                multihypothesis = []
                for detection in detections:
                    if np.array_equal(measurement_prediction.state_vector,
                                      detection.state_vector):
                        multihypothesis.append(
                            SingleProbabilityHypothesis(
                                prediction, detection,
                                measurement_prediction=measurement_prediction,
                                probability=0.9
                            ))
                        multihypothesis.append(
                            SingleProbabilityHypothesis(
                                prediction, MissedDetection(timestamp=time),
                                measurement_prediction=measurement_prediction,
                                probability=0.1
                            ))
                        break
                else:
                    multihypothesis.append(
                        SingleProbabilityHypothesis(
                            prediction, MissedDetection(timestamp=time),
                            measurement_prediction=measurement_prediction,
                            probability=0.1
                        ))
                associations[track] = MultipleHypothesis(multihypothesis)

            return associations
    return TestDataMixtureAssociator()


@pytest.fixture()
def updater():
    class TestUpdater:
        def update(self, hypothesis):
            return GaussianStateUpdate(hypothesis.measurement.state_vector,
                                       hypothesis.measurement.covar,
                                       hypothesis,
                                       0)
    return TestUpdater()



# -*- coding: utf-8 -*-
import datetime

import numpy as np
import pytest

from ...types.prediction import (
    StatePrediction, StateMeasurementPrediction,
    GaussianStatePrediction, GaussianMeasurementPrediction)


def test_stateprediction():
    """ StatePrediction test """

    with pytest.raises(TypeError):
        StatePrediction()

    state_vector = np.array([[-1.8513], [0.9994], [0], [0]]) * 1e4
    timestamp = datetime.datetime.now()

    state_prediction = StatePrediction(state_vector, timestamp)
    assert(np.array_equal(state_vector, state_prediction.state_vector))
    assert(timestamp == state_prediction.timestamp)


def test_statemeasurementprediction():
    """ MeasurementPrediction test """

    with pytest.raises(TypeError):
        StateMeasurementPrediction()

    state_vector = np.array([[-1.8513], [0.9994], [0], [0]]) * 1e4
    timestamp = datetime.datetime.now()

    measurement_prediction = StateMeasurementPrediction(state_vector,
                                                        timestamp)
    assert(np.array_equal(state_vector, measurement_prediction.state_vector))
    assert(timestamp == measurement_prediction.timestamp)


def test_gaussianstateprediction():
    """ GaussianStatePrediction test """

    with pytest.raises(TypeError):
        GaussianStatePrediction()

    mean = np.array([[-1.8513], [0.9994], [0], [0]]) * 1e4
    covar = np.array([[2.2128, 0, 0, 0],
                      [0.0002, 2.2130, 0, 0],
                      [0.3897, -0.00004, 0.0128, 0],
                      [0, 0.3897, 0.0013, 0.0135]]) * 1e3
    timestamp = datetime.datetime.now()

    with pytest.raises(TypeError):
        GaussianStatePrediction(mean)

    # Test state prediction
    state_prediction = GaussianStatePrediction(mean, covar, timestamp)
    assert(np.array_equal(mean, state_prediction.mean))
    assert(np.array_equal(covar, state_prediction.covar))
    assert(state_prediction.ndim == mean.shape[0])
    assert(state_prediction.timestamp == timestamp)


def test_gaussianmeasurementprediction():
    """ GaussianMeasurementPrediction test """

    with pytest.raises(TypeError):
        GaussianMeasurementPrediction()

    mean = np.array([[-1.8513], [0.9994], [0], [0]]) * 1e4
    covar = np.array([[2.2128, 0, 0, 0],
                      [0.0002, 2.2130, 0, 0],
                      [0.3897, -0.00004, 0.0128, 0],
                      [0, 0.3897, 0.0013, 0.0135]]) * 1e3
    cross_covar = np.array([[2.2128, 0, 0, 0],
                            [0.0002, 2.2130, 0, 0],
                            [0.3897, -0.00004, 0.0128, 0],
                            [0, 0.3897, 0.0013, 0.0135]]) * 1e3
    wrong_cross_covar = np.array([[2.2128, 0, 0],
                                  [0.0002, 2.2130, 0],
                                  [0.3897, -0.00004, 0.0128],
                                  [0, 0.3897, 0.0013]]) * 1e3
    timestamp = datetime.datetime.now()

    with pytest.raises(TypeError):
        GaussianMeasurementPrediction(mean)

    with pytest.raises(ValueError):
        GaussianMeasurementPrediction(mean, covar,
                                      cross_covar=wrong_cross_covar,
                                      timestamp=timestamp)

    # Test measurement prediction initiation without cross_covar
    measurement_prediction = GaussianMeasurementPrediction(
        mean, covar,
        timestamp=timestamp)
    assert(np.array_equal(mean, measurement_prediction.mean))
    assert(np.array_equal(covar, measurement_prediction.covar))
    assert(measurement_prediction.ndim == mean.shape[0])
    assert(measurement_prediction.timestamp == timestamp)
    assert(measurement_prediction.cross_covar is None)

    # Test measurement prediction initiation with cross_covar
    measurement_prediction = GaussianMeasurementPrediction(
        mean, covar,
        cross_covar=cross_covar,
        timestamp=timestamp)
    assert(np.array_equal(mean, measurement_prediction.mean))
    assert(np.array_equal(covar, measurement_prediction.covar))
    assert(np.array_equal(cross_covar, measurement_prediction.cross_covar))
    assert(measurement_prediction.ndim == mean.shape[0])
    assert(measurement_prediction.timestamp == timestamp)



# -*- coding: utf-8 -*-
import datetime

import numpy as np

from ...types.state import GaussianState
from ...types.track import Track
from ..error import CovarianceBasedDeleter


def test_cbd():
    """Test CovarianceBasedDeleter"""

    timestamp = datetime.datetime.now()
    state = GaussianState(
        np.array([[0], [0]]),
        np.array([[100, 0], [0, 1]]), timestamp)
    track = Track()
    track.append(state)
    tracks = {track}

    state = GaussianState(
        np.array([[0], [0]]),
        np.array([[1, 0], [0, 1]]), timestamp)
    track = Track()
    track.append(state)

    tracks.add(track)

    cover_deletion_thresh = 100
    deleter = CovarianceBasedDeleter(cover_deletion_thresh)

    deleted_tracks = deleter.delete_tracks(tracks)
    tracks -= deleted_tracks

    assert(len(tracks) == 1)
    assert(len(deleted_tracks) == 1)



# -*- coding: utf-8 -*-

import datetime

import pytest

from ..time import TimeRange


def test_timerange():
    # Test time range initialisation

    t1 = datetime.datetime(year=2018, month=3, day=1, hour=5, minute=3,
                           second=35, microsecond=500)
    t2 = datetime.datetime(year=2018, month=3, day=1, hour=6, minute=5,
                           second=41, microsecond=500)
    # Test creating without times
    with pytest.raises(TypeError):
        TimeRange()

    # Without start time
    with pytest.raises(TypeError):
        TimeRange(start_timestamp=t1)

    # Without end time
    with pytest.raises(TypeError):
        TimeRange(end_timestamp=t2)

    # Test an error is caught when end is after start
    with pytest.raises(ValueError):
        TimeRange(start_timestamp=t2, end_timestamp=t1)

    test_range = test_range = TimeRange(start_timestamp=t1, end_timestamp=t2)

    assert test_range.start_timestamp == t1
    assert test_range.end_timestamp == t2


def test_duration():
    # Test that duration is calculated properly
    t1 = datetime.datetime(year=2018, month=3, day=1, hour=5, minute=3,
                           second=35, microsecond=500)
    t2 = datetime.datetime(year=2018, month=3, day=1, hour=6, minute=5,
                           second=41, microsecond=500)
    test_range = TimeRange(start_timestamp=t1, end_timestamp=t2)

    assert test_range.duration == datetime.timedelta(seconds=3726)


def test_contains():
    # Test that timestamps are correctly determined to be in the range

    t1 = datetime.datetime(year=2018, month=3, day=1, hour=5, minute=3,
                           second=35, microsecond=500)
    t2 = datetime.datetime(year=2018, month=3, day=1, hour=6, minute=5,
                           second=41, microsecond=500)
    test_range = TimeRange(start_timestamp=t1, end_timestamp=t2)

    # Inside
    assert datetime.datetime(
        year=2018, month=3, day=1, hour=5, minute=10, second=3) in test_range

    # Outside

    assert not datetime.datetime(
        year=2018, month=3, day=1, hour=3, minute=10, second=3) in test_range

    # Lower edge
    assert t1 in test_range

    # Upper edge
    assert t2 in test_range



# -*- coding: utf-8 -*-
import numpy as np

from .base import Resampler
from ..types.numeric import Probability
from ..types.particle import Particle


class SystematicResampler(Resampler):

    def resample(self, particles):
        """Resample the particles

        Parameters
        ----------
        particles : list of :class:`~.Particle`
            The particles to be resampled according to their weight

        Returns
        -------
        particles : list of :class:`~.Particle`
            The resampled particles
        """

        n_particles = len(particles)
        weight = Probability(1/n_particles)
        cdf = np.cumsum([p.weight for p in particles])
        particles_listed = list(particles)
        # Pick random starting point
        u_i = np.random.uniform(0, 1 / n_particles)
        new_particles = []

        # Cycle through the cumulative distribution and copy the particle
        # that pushed the score over the current value
        for j in range(n_particles):

            u_j = u_i + (1 / n_particles) * j

            particle = particles_listed[np.argmax(u_j < cdf)]
            new_particles.append(
                Particle(particle.state_vector,
                         weight=weight,
                         parent=particle))

        return new_particles



# -*- coding: utf-8 -*-
"""Test for updater.kalman module"""
import pytest
import numpy as np

from ...models.measurement.linear import LinearGaussian
from ...types.detection import Detection
from ...types.hypothesis import SingleHypothesis
from ...types.prediction import (
    GaussianStatePrediction, GaussianMeasurementPrediction)
from ...types.state import GaussianState
from ...updater.kalman import (
    KalmanUpdater, ExtendedKalmanUpdater, UnscentedKalmanUpdater)


@pytest.mark.parametrize(
    "UpdaterClass, measurement_model, prediction, measurement",
    [
        (   # Standard Kalman
            KalmanUpdater,
            LinearGaussian(ndim_state=2, mapping=[0],
                           noise_covar=np.array([[0.04]])),
            GaussianStatePrediction(np.array([[-6.45], [0.7]]),
                                    np.array([[4.1123, 0.0013],
                                              [0.0013, 0.0365]])),
            Detection(np.array([[-6.23]]))
        ),
        (   # Extended Kalman
            ExtendedKalmanUpdater,
            LinearGaussian(ndim_state=2, mapping=[0],
                           noise_covar=np.array([[0.04]])),
            GaussianStatePrediction(np.array([[-6.45], [0.7]]),
                                    np.array([[4.1123, 0.0013],
                                              [0.0013, 0.0365]])),
            Detection(np.array([[-6.23]]))
        ),
        (   # Unscented Kalman
            UnscentedKalmanUpdater,
            LinearGaussian(ndim_state=2, mapping=[0],
                           noise_covar=np.array([[0.04]])),
            GaussianStatePrediction(np.array([[-6.45], [0.7]]),
                                    np.array([[4.1123, 0.0013],
                                              [0.0013, 0.0365]])),
            Detection(np.array([[-6.23]]))
        )
    ],
    ids=["standard", "extended", "unscented"]
)
def test_kalman(UpdaterClass, measurement_model, prediction, measurement):

    # Calculate evaluation variables
    eval_measurement_prediction = GaussianMeasurementPrediction(
        measurement_model.matrix()@prediction.mean,
        measurement_model.matrix()@prediction.covar
        @measurement_model.matrix().T
        + measurement_model.covar(),
        cross_covar=prediction.covar@measurement_model.matrix().T)
    kalman_gain = eval_measurement_prediction.cross_covar@np.linalg.inv(
        eval_measurement_prediction.covar)
    eval_posterior = GaussianState(
        prediction.mean
        + kalman_gain@(measurement.state_vector
                       - eval_measurement_prediction.mean),
        prediction.covar
        - kalman_gain@eval_measurement_prediction.covar@kalman_gain.T)

    # Initialise a kalman updater
    updater = UpdaterClass(measurement_model=measurement_model)

    # Get and assert measurement prediction
    measurement_prediction = updater.get_measurement_prediction(prediction)
    assert(np.allclose(measurement_prediction.mean,
                       eval_measurement_prediction.mean,
                       0, atol=1.e-14))
    assert(np.allclose(measurement_prediction.covar,
                       eval_measurement_prediction.covar,
                       0, atol=1.e-14))
    assert(np.allclose(measurement_prediction.cross_covar,
                       eval_measurement_prediction.cross_covar,
                       0, atol=1.e-14))

    # Perform and assert state update (without measurement prediction)
    posterior = updater.update(SingleHypothesis(
        prediction=prediction,
        measurement=measurement))
    assert(np.allclose(posterior.mean, eval_posterior.mean, 0, atol=1.e-14))
    assert(np.allclose(posterior.covar, eval_posterior.covar, 0, atol=1.e-14))
    assert(np.array_equal(posterior.hypothesis.prediction, prediction))
    assert (np.allclose(
        posterior.hypothesis.measurement_prediction.state_vector,
        measurement_prediction.state_vector, 0, atol=1.e-14))
    assert (np.allclose(posterior.hypothesis.measurement_prediction.covar,
                        measurement_prediction.covar, 0, atol=1.e-14))
    assert(np.array_equal(posterior.hypothesis.measurement, measurement))
    assert(posterior.timestamp == prediction.timestamp)

    # Perform and assert state update
    posterior = updater.update(SingleHypothesis(
        prediction=prediction,
        measurement=measurement,
        measurement_prediction=measurement_prediction))
    assert(np.allclose(posterior.mean, eval_posterior.mean, 0, atol=1.e-14))
    assert(np.allclose(posterior.covar, eval_posterior.covar, 0, atol=1.e-14))
    assert(np.array_equal(posterior.hypothesis.prediction, prediction))
    assert (np.allclose(
        posterior.hypothesis.measurement_prediction.state_vector,
        measurement_prediction.state_vector, 0, atol=1.e-14))
    assert (np.allclose(posterior.hypothesis.measurement_prediction.covar,
                        measurement_prediction.covar, 0, atol=1.e-14))
    assert(np.array_equal(posterior.hypothesis.measurement, measurement))
    assert(posterior.timestamp == prediction.timestamp)



# -*- coding: utf-8 -*-
import numpy as np

from .base import Tracker
from ..base import Property
from ..dataassociator import DataAssociator
from ..deleter import Deleter
from ..reader import DetectionReader
from ..initiator import Initiator
from ..updater import Updater
from ..types.update import GaussianStateUpdate
from ..functions import gm_reduce_single


class SingleTargetTracker(Tracker):
    """A simple single target tracker.

    Track a single object using Stone Soup components. The tracker works by
    first calling the :attr:`data_associator` with the active track, and then
    either updating the track state with the result of the :attr:`updater` if
    a detection is associated, or with the prediction if no detection is
    associated to the track. The track is then checked for deletion by the
    :attr:`deleter`, and if deleted the :attr:`initiator` is called to generate
    a new track. Similarly if no track is present (i.e. tracker is initialised
    or deleted in previous iteration), only the :attr:`initiator` is called.

    Parameters
    ----------

    Attributes
    ----------
    track : :class:`~.Track`
        Current track being maintained. Also accessible as the sole item in
        :attr:`tracks`
    """
    initiator = Property(
        Initiator,
        doc="Initiator used to initialise the track.")
    deleter = Property(
        Deleter,
        doc="Deleter used to delete the track")
    detector = Property(
        DetectionReader,
        doc="Detector used to generate detection objects.")
    data_associator = Property(
        DataAssociator,
        doc="Association algorithm to pair predictions to detections")
    updater = Property(
        Updater,
        doc="Updater used to update the track object to the new state.")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.track = None

    @property
    def tracks(self):
        if self.track is not None:
            return {self.track}
        else:
            return set()

    def tracks_gen(self):
        self.track = None

        for time, detections in self.detector.detections_gen():

            if self.track is not None:
                associations = self.data_associator.associate(
                        self.tracks, detections, time)
                if associations[self.track]:
                    state_post = self.updater.update(associations[self.track])
                    self.track.append(state_post)
                else:
                    self.track.append(
                        associations[self.track].prediction)

            if self.track is None or self.deleter.delete_tracks(self.tracks):
                new_tracks = self.initiator.initiate(detections)
                if new_tracks:
                    track = next(iter(new_tracks))
                    self.track = track
                else:
                    self.track = None

            yield time, self.tracks


class MultiTargetTracker(Tracker):
    """A simple multi target tracker.

    Track multiple objects using Stone Soup components. The tracker works by
    first calling the :attr:`data_associator` with the active tracks, and then
    either updating the track state with the result of the :attr:`updater` if
    a detection is associated, or with the prediction if no detection is
    associated to the track. Tracks are then checked for deletion by the
    :attr:`deleter`, and remaining unassociated detections are passed to the
    :attr:`initiator` to generate new tracks.

    Parameters
    ----------
    """
    initiator = Property(
        Initiator,
        doc="Initiator used to initialise the track.")
    deleter = Property(
        Deleter,
        doc="Initiator used to initialise the track.")
    detector = Property(
        DetectionReader,
        doc="Detector used to generate detection objects.")
    data_associator = Property(
        DataAssociator,
        doc="Association algorithm to pair predictions to detections")
    updater = Property(
        Updater,
        doc="Updater used to update the track object to the new state.")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracks = set()

    @property
    def tracks(self):
        return self._tracks.copy()

    def tracks_gen(self):
        self._tracks = set()

        for time, detections in self.detector.detections_gen():

            associations = self.data_associator.associate(
                self._tracks, detections, time)
            associated_detections = set()
            for track, hypothesis in associations.items():
                if hypothesis:
                    state_post = self.updater.update(hypothesis)
                    track.append(state_post)
                    associated_detections.add(hypothesis.measurement)
                else:
                    track.append(hypothesis.prediction)

            self._tracks -= self.deleter.delete_tracks(self._tracks)
            self._tracks |= self.initiator.initiate(
                detections - associated_detections)

            yield time, self.tracks


class MultiTargetMixtureTracker(Tracker):
    """A simple multi target tracker that receives associations from a
    (Guassian) Mixture associator.

    Track multiple objects using Stone Soup components. The tracker works by
    first calling the :attr:`data_associator` with the active tracks, and then
    either updating the track state with the result of the
    :attr:`data_associator` that reduces the (Gaussian) Mixture of all
    possible track-detection associations, or with the prediction if no
    detection is associated to the track. Tracks are then checked for deletion
    by the :attr:`deleter`, and remaining unassociated detections are passed
    to the :attr:`initiator` to generate new tracks.

    Parameters
    ----------
    """
    initiator = Property(
        Initiator,
        doc="Initiator used to initialise the track.")
    deleter = Property(
        Deleter,
        doc="Initiator used to initialise the track.")
    detector = Property(
        DetectionReader,
        doc="Detector used to generate detection objects.")
    data_associator = Property(
        DataAssociator,
        doc="Association algorithm to pair predictions to detections")
    updater = Property(
        Updater,
        doc="Updater used to update the track object to the new state.")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._tracks = set()

    @property
    def tracks(self):
        return self._tracks.copy()

    def tracks_gen(self):
        self._tracks = set()

        for time, detections in self.detector.detections_gen():

            associations = self.data_associator.associate(
                self._tracks, detections, time)
            unassociated_detections = set(detections)
            for track, multihypothesis in associations.items():

                # calculate each Track's state as a Gaussian Mixture of
                # its possible associations with each detection, then
                # reduce the Mixture to a single Gaussian State
                posterior_states = []
                posterior_state_weights = []
                for hypothesis in multihypothesis:
                    if not hypothesis:
                        posterior_states.append(hypothesis.prediction)
                    else:
                        posterior_states.append(
                            self.updater.update(hypothesis))
                    posterior_state_weights.append(
                        hypothesis.probability)

                means = np.array([state.state_vector for state
                                  in posterior_states])
                means = np.reshape(means, np.shape(means)[:-1])
                covars = np.array([state.covar for state
                                   in posterior_states])
                covars = np.reshape(covars, (np.shape(covars)))
                weights = np.array([weight for weight
                                    in posterior_state_weights])
                weights = np.reshape(weights, np.shape(weights))

                post_mean, post_covar = gm_reduce_single(means,
                                                         covars, weights)

                track.append(GaussianStateUpdate(
                    np.array(post_mean), np.array(post_covar),
                    multihypothesis,
                    multihypothesis[0].measurement.timestamp))

                # any detections in multihypothesis that had an
                # association score (weight) lower than or equal to the
                # association score of "MissedDetection" is considered
                # unassociated - candidate for initiating a new Track
                missed_detection_weight = next(
                    hyp.weight for hyp in multihypothesis if not hyp)

                for hyp in multihypothesis:
                    if hyp.weight > missed_detection_weight:
                        if hyp.measurement in unassociated_detections:
                            unassociated_detections.remove(hyp.measurement)

            self._tracks -= self.deleter.delete_tracks(self._tracks)
            self._tracks |= self.initiator.initiate(
                unassociated_detections)

            yield time, self.tracks



# -*- coding: utf-8 -*-
"""Reader classes are used for getting data into the framework."""
from .base import Reader, DetectionReader, GroundTruthReader, SensorDataReader

__all__ = [
    'Reader', 'DetectionReader', 'GroundTruthReader', 'SensorDataReader']



# coding: utf-8

import datetime

import scipy as sp
from scipy.stats import multivariate_normal

from ..linear import (
    ConstantAcceleration, CombinedLinearGaussianTransitionModel)


def test_cam1dodel():
    """ ConstantAcceleration Transition Model test """
    state_vec = sp.array([[3.0], [1.0], [0.1]])
    noise_diff_coeffs = sp.array([[0.01]])
    base(state_vec, noise_diff_coeffs)


def test_ca2dmodel():
    """ ConstantAcceleration2D Transition Model test """
    state_vec = sp.array([[3.0], [1.0], [0.1],
                          [2.0], [2.0], [0.2]])
    noise_diff_coeffs = sp.array([0.01, 0.02])
    base(state_vec, noise_diff_coeffs)


def test_ca3dmodel():
    """ ConstantAcceleration3D Transition Model test """
    state_vec = sp.array([[3.0], [1.0], [0.1],
                          [2.0], [2.0], [0.2],
                          [4.0], [0.5], [0.05]])
    noise_diff_coeffs = sp.array([0.01, 0.02, 0.005])
    base(state_vec, noise_diff_coeffs)


def base(state_vec, noise_diff_coeffs):
    """ Base test for n-dimensional ConstantAcceleration Transition Models """

    # Create a 1D ConstantAcceleration or an n-dimensional
    # CombinedLinearGaussianTransitionModel object
    dim = len(state_vec) // 3  # pos, vel, acc for each dimension
    if dim == 1:
        model_obj = ConstantAcceleration(noise_diff_coeff=noise_diff_coeffs[0])
    else:
        model_list = [ConstantAcceleration(
            noise_diff_coeff=noise_diff_coeffs[i]) for i in range(0, dim)]
        model_obj = CombinedLinearGaussianTransitionModel(model_list)

    # State related variables
    state_vec = state_vec
    old_timestamp = datetime.datetime.now()
    timediff = 1  # 1sec
    new_timestamp = old_timestamp + datetime.timedelta(seconds=timediff)
    time_interval = new_timestamp - old_timestamp

    # Model-related components
    noise_diff_coeffs = noise_diff_coeffs  # m/s^3
    base_mat = sp.array([[1, timediff, sp.power(timediff, 2)],
                         [0, 1, timediff],
                         [0, 0, 1]])
    mat_list = [base_mat for num in range(0, dim)]
    F = sp.linalg.block_diag(*mat_list)

    base_covar = sp.array([[sp.power(timediff, 5) / 20,
                            sp.power(timediff, 4) / 8,
                            sp.power(timediff, 3) / 6],
                           [sp.power(timediff, 4) / 8,
                            sp.power(timediff, 3) / 3,
                            sp.power(timediff, 2) / 2],
                           [sp.power(timediff, 3) / 6,
                            sp.power(timediff, 2) / 2,
                            timediff]])
    covar_list = [base_covar*sp.power(noise_diff_coeffs[i], 2)
                  for i in range(0, dim)]
    Q = sp.linalg.block_diag(*covar_list)

    # Ensure ```model_obj.transfer_function(time_interval)``` returns F
    assert sp.array_equal(F, model_obj.matrix(
        timestamp=new_timestamp, time_interval=time_interval))

    # Ensure ```model_obj.covar(time_interval)``` returns Q
    assert sp.array_equal(Q, model_obj.covar(
        timestamp=new_timestamp, time_interval=time_interval))

    # Propagate a state vector throught the model
    # (without noise)
    new_state_vec_wo_noise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=0)
    assert sp.array_equal(new_state_vec_wo_noise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (without noise)
    prob = model_obj.pdf(new_state_vec_wo_noise,
                         state_vec,
                         timestamp=new_timestamp,
                         time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_wo_noise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throughout the model
    # (with internal noise)
    new_state_vec_w_inoise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval)
    assert not sp.array_equal(new_state_vec_w_inoise, F@state_vec)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model_obj.pdf(new_state_vec_w_inoise,
                         state_vec,
                         timestamp=new_timestamp,
                         time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_inoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)

    # Propagate a state vector throught the model
    # (with external noise)
    noise = model_obj.rvs(timestamp=new_timestamp, time_interval=time_interval)
    new_state_vec_w_enoise = model_obj.function(
        state_vec,
        timestamp=new_timestamp,
        time_interval=time_interval,
        noise=noise)
    assert sp.array_equal(new_state_vec_w_enoise, F@state_vec+noise)

    # Evaluate the likelihood of the predicted state, given the prior
    # (with noise)
    prob = model_obj.pdf(new_state_vec_w_enoise, state_vec,
                         timestamp=new_timestamp, time_interval=time_interval)
    assert sp.array_equal(prob, multivariate_normal.pdf(
        new_state_vec_w_enoise.T,
        mean=sp.array(F@state_vec).ravel(),
        cov=Q).T)



# -*- coding: utf-8 -*-
from abc import abstractmethod

import numpy as np
from scipy.spatial import distance

from .base import Base, Property


class Measure(Base):
    """Measure base type

    A measure provides a means to assess the seperation between two
    :class:`~.State` objects state1 and state2.
    """
    mapping = Property(
        np.array,
        default=None,
        doc="Mapping array which specifies which elements within the"
            " state vectors are to be assessed as part of the measure"
    )

    @abstractmethod
    def __call__(self, state1, state2):
        r"""
        Compute the distance between a pair of :class:`~.State` objects

        Parameters
        ----------
        state1 : :class:`~.State`
        state2 : :class:`~.State`

        Returns
        -------
        float
            distance measure between a pair of input :class:`~.State` objects

        """
        return NotImplementedError


class Euclidean(Measure):
    r"""Euclidean distance measure

    This measure returns the Euclidean distance between a pair of
    :class:`~.State` objects.

    The Euclidean distance between a pair of state vectors :math:`u` and
    :math:`v` is defined as:

    .. math::
         \sqrt{\sum_{i=1}^{N}{(u_i - v_i)^2}}

    """
    def __call__(self, state1, state2):
        r"""Calculate the Euclidean distance between a pair of state vectors

        Parameters
        ----------
        state1 : :class:`~.State`
        state2 : :class:`~.State`

        Returns
        -------
        float
            Euclidean distance between two input :class:`~.State`

        """
        # Calculate Euclidean distance between two state
        if self.mapping is not None:
            return distance.euclidean(state1.state_vector[self.mapping],
                                      state2.state_vector[self.mapping])
        else:
            return distance.euclidean(state1.state_vector, state2.state_vector)


class EuclideanWeighted(Measure):
    r"""Weighted Euclidean distance measure

    This measure returns the Euclidean distance between a pair of
    :class:`~.State` objects, taking into account a specified weighting.

    The Weighted Euclidean distance between a pair of state vectors :math:`u`
    and :math:`v` with weighting :math:`w` is defined as:

    .. math::
       \sqrt{\sum_{i=1}^{N}{w_i|(u_i - v_i)^2}}

    Note
    ----
    The EuclideanWeighted object has a property called weighting, which
    allows the method to be called on different pairs of states.
    If different weightings need to be used then multiple
    :class:`Measure` objects must be created with the specific weighting

    """
    weighting = Property(
        [np.array],
        doc="Weighting vector for the Euclidean calculation")

    def __call__(self, state1, state2):
        r"""Calculate the weighted Euclidean distance between a pair of state
        objects

        Parameters
        ----------
        state1 : :class:`~.State`
        state2 : :class:`~.State`

        Returns
        -------
        dist : float
            Weighted euclidean distance between two input
            :class:`~.State` objects

        """
        if self.mapping is not None:
            return distance.euclidean(state1.state_vector[self.mapping],
                                      state2.state_vector[self.mapping],
                                      self.weighting)
        else:
            return distance.euclidean(state1.state_vector,
                                      state2.state_vector,
                                      self.weighting)


class Mahalanobis(Measure):
    r"""Mahalanobis distance measure

    This measure returns the Mahalanobis distance between a pair of
    :class:`~.State` objects taking into account the distribution (i.e.
    the :class:`~.CovarianceMatrix`) of the first :class:`.State` object

    The Mahalanobis distance between a distribution with mean :math:`\mu` and
    Covariance matrix :math:`\Sigma` and a point :math:`x` is defined as:

    .. math::
            \sqrt{( {\mu - x})  \Sigma^{-1}  ({\mu - x}^T )}


    """
    def __call__(self, state1, state2):
        r"""Calculate the Mahalanobis distance between a pair of state objects

        Parameters
        ----------
        state1 : :class:`~.State`
        state2 : :class:`~.State`

        Returns
        -------
        float
            Mahalanobis distance between a pair of input :class:`~.State`
            objects

        """
        if self.mapping is not None:
            u = state1.state_vector[self.mapping]
            v = state2.state_vector[self.mapping]
            # extract the mapped covariance data
            rows = np.array(self.mapping, dtype=np.intp)
            columns = np.array(self.mapping, dtype=np.intp)
            cov = state1.covar[rows[:, np.newaxis], columns]
        else:
            u = state1.state_vector
            v = state2.state_vector
            cov = state1.covar

        vi = np.linalg.inv(cov)

        return distance.mahalanobis(u, v, vi)


class SquaredGaussianHellinger(Measure):
    r"""Squared Gaussian Hellinger distance measure

    This measure returns the Squared Hellinger distance between a pair of
    :class:`~.GaussianState` multivariate objects.

    The Squared Hellinger distance between two multivariate normal
    distributions :math:`P \sim N(\mu_1,\Sigma_1)` and
    :math:`Q \sim N(\mu_2,\Sigma_2)` is defined as:

    .. math::
            1 - \sqrt{\frac{det(\Sigma_1)^{1/4}det(\Sigma_2)^{1/4}}
            {det(\Sigma_1+\Sigma_2/2)^{1/2}}}
            exp\bigg(\frac{-1}{8}(\mu_1-\mu_2)^T
            (\frac{\Sigma_1+\Sigma_2}{2})^{-1}(\mu_1-\mu_2)\bigg)

    Note
    ----
    This distance is bounded between 0 and :math:`\sqrt{2}`
    """
    def __call__(self, state1, state2):
        r""" Calculate the Squared Hellinger distance multivariate normal
        distributions

        Parameters
        ----------
        state1 : :class:`~.GaussianState`
        state2 : :class:`~.GaussianState`

        Returns
        -------
        float
            Squared Hellinger distance between two input
            :class:`~.GaussianState`



        """
        if self.mapping is not None:
            mu1 = state1.state_vector[self.mapping]
            mu2 = state2.state_vector[self.mapping]

            # extract the mapped covariance data
            rows = np.array(self.mapping, dtype=np.intp)
            columns = np.array(self.mapping, dtype=np.intp)
            sigma1 = state1.covar[rows[:, np.newaxis], columns]
            sigma2 = state2.covar[rows[:, np.newaxis], columns]
        else:
            mu1 = state1.state_vector
            mu2 = state2.state_vector
            sigma1 = state1.covar
            sigma2 = state2.covar

        sigma1_plus_sigma2 = sigma1 + sigma2
        if np.linalg.det(sigma1_plus_sigma2) > 1e-10:
            mu1_minus_mu2 = mu1 - mu2
            E = mu1_minus_mu2.T*(np.linalg.inv(sigma1_plus_sigma2)
                                 / mu1_minus_mu2)
            epsilon = -0.125*E
            numerator = np.sqrt(np.linalg.det(sigma1*sigma2))
            denominator = np.linalg.det(sigma1_plus_sigma2/2)
            squared_hellinger = 1 - (
                np.sqrt(numerator/denominator)*np.exp(epsilon))
        else:
            mu1_minus_mu2 = mu1 - mu2
            temp = mu1_minus_mu2.T * mu1_minus_mu2
            squared_hellinger = temp
        squared_hellinger = squared_hellinger[0, 0]
        return squared_hellinger


class GaussianHellinger(SquaredGaussianHellinger):
    r"""Gaussian Hellinger distance measure

    This measure returns the Hellinger distance between a pair of
    :class:`~.GaussianState` multivariate objects.

    The Hellinger distance between two multivariate normal distributions
    :math:`P \sim N(\mu_1,\Sigma_1)` and :math:`Q \sim N(\mu_2,\Sigma_2)`
    is defined as:

    .. math::
            \sqrt{1 - \sqrt{\frac{det(\Sigma_1)^{1/4}det(\Sigma_2)^{1/4}}
            {det(\Sigma_1+\Sigma_2/2)^{1/2}}}
            exp\bigg(\frac{-1}{8}(\mu_1-\mu_2)^T
            (\frac{\Sigma_1+\Sigma_2}{2})^{-1}(\mu_1-\mu_2)\bigg)}

    Note
    ----
    This distance is bounded between 0 and 1
    """
    def __call__(self, state1, state2):
        r""" Calculate the Hellinger distance between 2 state elements

        Parameters
        ----------
        state1 : :class:`~.GaussianState`
        state2 : :class:`~.GaussianState`

        Returns
        -------
        float
            Hellinger distance between two input :class:`~.GaussianState`


        """
        return np.sqrt(super().__call__(state1, state2))



# -*- coding: utf-8 -*-
import pytest

from ...types.prediction import (
    GaussianMeasurementPrediction, GaussianStatePrediction)


@pytest.fixture()
def predictor():
    class TestGaussianPredictor:
        def predict(self, prior, control_input=None, timestamp=None, **kwargs):
            return GaussianStatePrediction(prior.state_vector + 1,
                                           prior.covar * 2, timestamp)
    return TestGaussianPredictor()


@pytest.fixture()
def updater():
    class TestGaussianUpdater:
        def get_measurement_prediction(self, state_prediction,
                                       measurement_model=None, **kwargs):
            return GaussianMeasurementPrediction(state_prediction.state_vector,
                                                 state_prediction.covar,
                                                 state_prediction.timestamp)
    return TestGaussianUpdater()



# -*- coding: utf-8 -*-
import datetime

import numpy as np

from ...functions import cart2pol
from ...types.angle import Bearing
from ...types.array import StateVector, CovarianceMatrix
from ...types.state import State
from ..radar import RadarRangeBearing, RadarRotatingRangeBearing


def h2d(state_vector, translation_offset, rotation_offset):

    xyz = [[state_vector[0, 0] - translation_offset[0, 0]],
           [state_vector[1, 0] - translation_offset[1, 0]],
           [0]]

    # Get rotation matrix
    theta_z = - rotation_offset[2, 0]
    cos_z, sin_z = np.cos(theta_z), np.sin(theta_z)
    rot_z = np.array([[cos_z, -sin_z, 0],
                      [sin_z, cos_z, 0],
                      [0, 0, 1]])

    theta_y = - rotation_offset[1, 0]
    cos_y, sin_y = np.cos(theta_y), np.sin(theta_y)
    rot_y = np.array([[cos_y, 0, sin_y],
                      [0, 1, 0],
                      [-sin_y, 0, cos_y]])

    theta_x = - rotation_offset[0, 0]
    cos_x, sin_x = np.cos(theta_x), np.sin(theta_x)
    rot_x = np.array([[1, 0, 0],
                      [0, cos_x, -sin_x],
                      [0, sin_x, cos_x]])

    rotation_matrix = rot_z@rot_y@rot_x

    xyz_rot = rotation_matrix @ xyz
    x = xyz_rot[0, 0]
    y = xyz_rot[1, 0]
    # z = 0  # xyz_rot[2, 0]

    rho = np.sqrt(x**2 + y**2)
    phi = np.arctan2(y, x)

    return np.array([[Bearing(phi)], [rho]])


def test_simple_radar():

    # Input arguments
    # TODO: pytest parametarization
    noise_covar = CovarianceMatrix(np.array([[0.015, 0],
                                             [0, 0.1]]))
    radar_position = StateVector(
        np.array(([[1], [1]])))
    radar_orientation = StateVector([[0], [0], [0]])
    target_state = State(radar_position +
                         np.array([[1], [1]]),
                         timestamp=datetime.datetime.now())
    measurement_mapping = np.array([0, 1])

    # Create a radar object
    radar = RadarRangeBearing(
        position=radar_position,
        orientation=radar_orientation,
        ndim_state=2,
        mapping=measurement_mapping,
        noise_covar=noise_covar)

    # Assert that the object has been correctly initialised
    assert(np.equal(radar.position, radar_position).all())
    assert(np.equal(radar.measurement_model.translation_offset,
                    radar_position).all())

    # Generate a noiseless measurement for the given target
    measurement = radar.gen_measurement(target_state, noise=0)
    rho, phi = cart2pol(target_state.state_vector[0, 0]
                        - radar_position[0, 0],
                        target_state.state_vector[1, 0]
                        - radar_position[1, 0])

    # Assert correction of generated measurement
    assert(measurement.timestamp == target_state.timestamp)
    assert(np.equal(measurement.state_vector,
                    StateVector(np.array([[phi], [rho]]))).all())


def test_rotating_radar():

    # Input arguments
    # TODO: pytest parametarization
    timestamp = datetime.datetime.now()
    noise_covar = CovarianceMatrix(np.array([[0.015, 0],
                                             [0, 0.1]]))

    # The radar is positioned at (1,1)
    radar_position = StateVector(
        np.array(([[1], [1]])))
    # The radar is facing left/east
    radar_orientation = StateVector([[0], [0], [np.pi]])
    # The radar antenna is facing opposite the radar orientation
    dwell_center = State(StateVector([[-np.pi]]),
                         timestamp=timestamp)
    rpm = 20            # 20 Rotations Per Minute
    max_range = 100     # Max range of 100m
    fov_angle = np.pi/3       # FOV angle of pi/3

    target_state = State(radar_position +
                         np.array([[5], [5]]),
                         timestamp=timestamp)
    measurement_mapping = np.array([0, 1])

    # Create a radar object
    radar = RadarRotatingRangeBearing(
        position=radar_position,
        orientation=radar_orientation,
        ndim_state=2,
        mapping=measurement_mapping,
        noise_covar=noise_covar,
        dwell_center=dwell_center,
        rpm=rpm,
        max_range=max_range,
        fov_angle=fov_angle)

    # Assert that the object has been correctly initialised
    assert(np.equal(radar.position, radar_position).all())
    assert(np.equal(radar.measurement_model.translation_offset,
                    radar_position).all())

    # Generate a noiseless measurement for the given target
    measurement = radar.gen_measurement(target_state, noise=0)

    # Assert measurement is None since target is not in FOV
    assert(measurement is None)

    # Rotate radar such that the target is in FOV
    timestamp = timestamp + datetime.timedelta(seconds=0.5)
    target_state = State(radar_position +
                         np.array([[5], [5]]),
                         timestamp=timestamp)
    measurement = radar.gen_measurement(target_state, noise=0)
    eval_m = h2d(target_state.state_vector,
                 radar.position,
                 radar.orientation+[[0],
                                    [0],
                                    [radar.dwell_center.state_vector[0, 0]]])

    # Assert correction of generated measurement
    assert(measurement.timestamp == target_state.timestamp)
    assert(np.equal(measurement.state_vector, eval_m).all())



# -*- coding: utf-8 -*-
"""Simulators for data input into Stone Soup.

Stone Soup can make use of simulators to generate ground truth tracks, sensor
data and detections. These are similar to :class:`reader.Reader`, but data is
generated, rather than read from a file, etc. They should come with various
configuration options to allow customisation of the simulation.
"""

from .base import (
    Simulator, DetectionSimulator, GroundTruthSimulator, SensorSimulator)

__all__ = ['Simulator', 'DetectionSimulator', 'GroundTruthSimulator',
           'SensorSimulator']



# -*- coding: utf-8 -*-
"""Contains collection of error based deleters"""
import numpy as np

from ..base import Property
from .base import Deleter


class CovarianceBasedDeleter(Deleter):
    """ Track deleter based on covariance matrix size.

    Deletes tracks whose state covariance matrix (more specifically its trace)
    exceeds a given threshold.
    """

    covar_trace_thresh = Property(
        float, doc="Covariance matrix trace threshold")

    def check_for_deletion(self, track, **kwargs):
        """Check if a given track should be deleted

        A track is flagged for deletion if the trace of its state covariance
        matrix is higher than :py:attr:`~covar_trace_thresh`.

        Parameters
        ----------
        track : :class:`stonesoup.types.Track`
            A track object to be checked for deletion.

        Returns
        -------
        : :class:`bool`
            ``True`` if track should be deleted, ``False`` otherwise.
        """

        track_covar_trace = np.trace(track.state.covar)

        if(track_covar_trace > self.covar_trace_thresh):
            return True
        return False



# -*- coding: utf-8 -*-
"""Base classes for Stone Soup Predictor interface"""
from abc import abstractmethod

from ..base import Base, Property
from ..models.transition import TransitionModel
from ..models.control import ControlModel


class Predictor(Base):
    """Predictor base class

    A predictor is used to advance a state to another point in time, by
    utilising a specified :class:`~.TransitionModel`. In addition a
    :class:`~.ControlModel` may be used to model an external influence to the
    state.
    """

    transition_model = Property(TransitionModel, doc="transition model")
    control_model = Property(ControlModel, default=None, doc="control model")

    @abstractmethod
    def predict(self, prior, control_input=None, timestamp=None, **kwargs):
        """Predict state

        Parameters
        ----------
        prior : :class:`~.State`
            State
        control_input : :class:`~.State`
            State
        timestamp : :class:`datetime.datetime`
            Time which to predict to which will be passed to transition model

        Returns
        -------
        : :class:`~.StatePrediction`
            State prediction
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

import scipy as sp
from scipy.stats import multivariate_normal

from ...base import Property
from ...types.array import CovarianceMatrix
from ..base import LinearModel, GaussianModel
from .base import MeasurementModel


class LinearGaussian(MeasurementModel, LinearModel, GaussianModel):
    r"""This is a class implementation of a time-invariant 1D
    Linear-Gaussian Measurement Model.

    The model is described by the following equations:

    .. math::

      y_t = H_k*x_t + v_k,\ \ \ \   v(k)\sim \mathcal{N}(0,R)

    where ``H_k`` is a (:py:attr:`~ndim_meas`, :py:attr:`~ndim_state`) \
    matrix and ``v_k`` is Gaussian distributed.

    """

    noise_covar = Property(CovarianceMatrix, doc="Noise covariance")

    @property
    def ndim_meas(self):
        """ndim_meas getter method

        Returns
        -------
        :class:`int`
            The number of measurement dimensions
        """

        return len(self.mapping)

    def matrix(self, **kwargs):
        """Model matrix :math:`H(t)`

        Returns
        -------
        :class:`numpy.ndarray` of shape \
        (:py:attr:`~ndim_meas`, :py:attr:`~ndim_state`)
            The model matrix evaluated given the provided time interval.
        """

        model_matrix = sp.zeros((self.ndim_meas, self.ndim_state))
        for dim_meas, dim_state in enumerate(self.mapping):
            if dim_state is not None:
                model_matrix[dim_meas, dim_state] = 1

        return model_matrix

    def function(self, state_vector, noise=None, **kwargs):
        """Model function :math:`h(t,x(t),w(t))`

        Parameters
        ----------
        state_vector: :class:`~.StateVector`
            An input state vector
        noise: :class:`numpy.ndarray`
            An externally generated random process noise sample (the default in
            `None`, in which case process noise will be generated internally)

        Returns
        -------
        :class:`numpy.ndarray` of shape (:py:attr:`~ndim_meas`, 1)
            The model function evaluated given the provided time interval.
        """

        if noise is None:
            noise = self.rvs()

        return self.matrix(**kwargs)@state_vector + noise

    def covar(self, **kwargs):
        """Returns the measurement model noise covariance matrix.

        Returns
        -------
        :class:`~.CovarianceMatrix` of shape\
        (:py:attr:`~ndim_meas`, :py:attr:`~ndim_meas`)
            The measurement noise covariance.
        """

        return self.noise_covar

    def rvs(self, num_samples=1, **kwargs):
        r""" Model noise/sample generation function

        Generates noise samples from the measurement model.

        In mathematical terms, this can be written as:

        .. math::

            v_t \sim \mathcal{N}(0,R_t)

        where :math:`v_t =` ``noise``.

        Parameters
        ----------
        num_samples: scalar, optional
            The number of samples to be generated (the default is 1)

        Returns
        -------
        noise : 2-D array of shape (:py:attr:`~ndim_meas`, ``num_samples``)
            A set of Np samples, generated from the model's noise
            distribution.
        """

        noise = multivariate_normal.rvs(
            sp.zeros(self.ndim_meas), self.covar(), num_samples)

        if num_samples == 1:
            return noise.reshape((-1, 1))
        else:
            return noise.T

    def pdf(self, meas_vec, state_vec, **kwargs):
        r""" Measurement pdf/likelihood evaluation function

        Evaluates the pdf/likelihood of the (set of) measurement vector(s)
        ``meas_vec``, given the (set of) state vector(s) ``state_vec``.

        In mathematical terms, this can be written as:

        .. math::

            p(y_t | x_t) = \mathcal{N}(y_t; x_t, R_t)

        Parameters
        ----------
        meas_vec : :class:`~.StateVector`
            A measurement
        state_vec : :class:`~.StateVector`
            A state

        Returns
        -------
        :class:`float`
            The likelihood of ``meas``, given ``state``
        """

        likelihood = multivariate_normal.pdf(
            meas_vec.T,
            mean=(self.matrix()@state_vec).ravel(),
            cov=self.covar()
        )
        return likelihood



# -*- coding: utf-8 -*-
import datetime
from textwrap import dedent

import numpy as np

from ..generic import CSVDetectionReader


def test_csv(tmpdir):
    csv_filename = tmpdir.join("test.csv")
    with csv_filename.open('w') as csv_file:
        csv_file.write(dedent("""\
            x,y,z,identifier,t
            10,20,30,22018332,2018-01-01T14:00:00Z
            11,21,31,22018332,2018-01-01T14:01:00Z
            12,22,32,22018332,2018-01-01T14:02:00Z
            """))

    # run test with:
    #   - 'metadata_fields' for 'CSVDetectionReader' == default
    #   - copy all metadata items
    csv_reader = CSVDetectionReader(csv_filename.strpath, ["x", "y"], "t")
    detections = [
        detection
        for _, detections in csv_reader.detections_gen()
        for detection in detections]

    for n, detection in enumerate(detections):
        assert np.array_equal(
            detection.state_vector, np.array([[10 + n], [20 + n]]))
        assert detection.timestamp.hour == 14
        assert detection.timestamp.minute == n
        assert detection.timestamp.date() == datetime.date(2018, 1, 1)
        assert isinstance(detection.metadata, dict)
        assert len(detection.metadata) == 2
        assert 'z' in detection.metadata.keys()
        assert 'identifier' in detection.metadata.keys()
        assert int(detection.metadata['z']) == 30 + n
        assert detection.metadata['identifier'] == '22018332'

    # run test with:
    #   - 'metadata_fields' for 'CSVDetectionReader' contains
    #       'z' but not 'identifier'
    #   - 'time_field_format' is specified
    csv_reader = CSVDetectionReader(csv_filename.strpath, ["x", "y"], "t",
                                    time_field_format="%Y-%m-%dT%H:%M:%SZ",
                                    metadata_fields=["z"])
    detections = [
        detection
        for _, detections in csv_reader.detections_gen()
        for detection in detections]

    for n, detection in enumerate(detections):
        assert np.array_equal(
            detection.state_vector, np.array([[10 + n], [20 + n]]))
        assert detection.timestamp.hour == 14
        assert detection.timestamp.minute == n
        assert detection.timestamp.date() == datetime.date(2018, 1, 1)
        assert isinstance(detection.metadata, dict)
        assert len(detection.metadata) == 1
        assert 'z' in detection.metadata.keys()
        assert int(detection.metadata['z']) == 30 + n

    # run test with:
    #   - 'metadata_fields' for 'CSVDetectionReader' contains
    #       column names that do not exist in CSV file
    #   - 'time' field represented as a Unix epoch timestamp
    with csv_filename.open('w') as csv_file:
        csv_file.write(dedent("""\
            x,y,z,identifier,t
            10,20,30,22018332,1514815200
            11,21,31,22018332,1514815260
            12,22,32,22018332,1514815320
            """))

    csv_reader = CSVDetectionReader(csv_filename.strpath, ["x", "y"], "t",
                                    metadata_fields=["heading"],
                                    timestamp=True)
    detections = [
        detection
        for _, detections in csv_reader.detections_gen()
        for detection in detections]

    for n, detection in enumerate(detections):
        assert np.array_equal(
            detection.state_vector, np.array([[10 + n], [20 + n]]))
        assert detection.timestamp.hour == 14
        assert detection.timestamp.minute == n
        assert detection.timestamp.date() == datetime.date(2018, 1, 1)
        assert isinstance(detection.metadata, dict)
        assert len(detection.metadata) == 0



# -*- coding: utf-8 -*-
import numpy as np
from math import cos, sin
from scipy.linalg import expm

from ..base import Property
from ..types.state import StateVector
from ..sensor import Sensor
from ..functions import cart2pol
from .base import Platform


class SensorPlatform(Platform):
    """A simple Platform that can carry a number of different sensors and is
    capable of moving based upon the :class:`~.TransitionModel`.

    The location of platform mounted sensors will be maintained relative to
    the sensor position. Simple platforms move within a 2 or 3 dimensional
    rectangular cartesian space.

    A simple platform is considered to always be aligned with its principle
    velocity. It does not take into account issues such as bank angle or body
    deformation (e.g. flex).

    """

    sensors = Property([Sensor], doc="A list of N mounted sensors")
    mounting_offsets = Property(
        [np.array], doc="A list of sensor offsets (For now expressed\
                            as a Nxn array of nD Cartesian coordinates)")
    mounting_mappings = Property(
        [np.array], doc="Mappings between the platform state vector and the\
                            individuals sensors mounting offset (For now\
                            expressed as a nxN array of nD Cartesian\
                            coordinates or a 1xN array where a single\
                            mapping will be applied to all sensors)")

    # TODO: Determine where a platform coordinate frame should be maintained

    def __init__(self, *args, **kwargs):
        """
        Ensure that the platform location and the sensor locations are
        consistent at initialisation.
        """
        super().__init__(*args, **kwargs)
        if self.mounting_mappings.max() > len(self.state.state_vector):
            raise IndexError(
                "Platform state vector length and sensor mounting mapping "
                "are incompatible")

        if len(self.sensors) != self.mounting_offsets.shape[0]:
            raise IndexError(
                "Number of sensors associated with the platform does not "
                "match the number of sensor mounting offsets specified")

        if ((len(self.sensors) != self.mounting_mappings.shape[0]) and
                (self.mounting_mappings.shape[0] != 1)):
            raise IndexError(
                "Number of sensors associated with the platform does not "
                "match the number of mounting mappings specified")

        if ((self.mounting_mappings.shape[0] == 1) and
                len(self.sensors) > 1):
            mapping_array = np.empty((0, self.mounting_mappings.shape[1]), int)
            for i in range(len(self.sensors)):
                mapping_array = np.append(mapping_array,
                                          self.mounting_mappings,
                                          axis=0)
            self.mounting_mappings = mapping_array

        self._move_sensors()

    # TODO: create add_sensor method

    def move(self, timestamp=None, **kwargs):
        """Propagate the platform position using the :attr:`transition_model`,
        and use _move_sensors method to update sensor positions, this in turn
        calls _get_rotated_offset to modify sensor offsets relative to the
        platforms velocity vector

        Parameters
        ----------
        timestamp: :class:`datetime.datetime`, optional
            A timestamp signifying when the maneuver completes
            (the default is `None`)

        Notes
        -----
        This methods updates the value of :attr:`position` and :attr:`sensors`
        """
        # Call superclass method to update platform state
        super().move(timestamp=timestamp, **kwargs)
        # Move the platforms sensors relative to the platform
        self._move_sensors()

    def _move_sensors(self):
        """ Propogate the Sensor positions based upon the mounting
        offsets and the platform position and heading post manoeuvre.

        Notes
        -----
        Method assumes that if a platform has a transition model it will have
        velocity components. A sensor offset will therefore be rotated based
        upon the platforms velocity (i.e. direction of motion)
        """

        # Update the positions of all sensors relative to the platform
        for i in range(len(self.sensors)):
            if (hasattr(self, 'transition_model') &
                    (np.absolute(self.state.state_vector[
                        self.mounting_mappings[0]+1]).max() > 0)):
                new_sensor_pos = self._get_rotated_offset(i)
                for j in range(self.mounting_offsets.shape[1]):
                    new_sensor_pos[j] = new_sensor_pos[j] + \
                        (self.state.state_vector[
                            self.mounting_mappings[i, j]])
            else:
                new_sensor_pos = np.zeros([self.mounting_offsets.shape[1], 1])
                for j in range(self.mounting_offsets.shape[1]):
                    new_sensor_pos[j] = (self.state.state_vector[
                        self.mounting_mappings[i, j]] +
                        self.mounting_offsets[i, j])
            self.sensors[i].set_position(StateVector(new_sensor_pos))
            vel = np.zeros([self.mounting_mappings.shape[1], 1])
            for j in range(self.mounting_mappings.shape[1]):
                vel[j, 0] = self.state.state_vector[
                    self.mounting_mappings[i, j] + 1]
            abs_vel, heading = cart2pol(vel[0, 0], vel[1, 0])
            self.sensors[i].set_orientation(
                StateVector([[0], [0], [heading]]))

    def _get_rotated_offset(self, i):
        """ Determine the sensor mounting offset for the platforms relative
        orientation.

        Parameters
        ----------
        i : int
            Integer reference to the sensor index

        Returns
        -------
        np.array
            Sensor mounting offset rotated relative to platform motion
        """

        vel = np.zeros([self.mounting_mappings.shape[1], 1])
        for j in range(self.mounting_mappings.shape[1]):
            vel[j, 0] = self.state.state_vector[
                self.mounting_mappings[i, j] + 1]

        rot = _get_rotation_matrix(vel)
        return np.transpose(np.dot(rot, self.mounting_offsets[i])[np.newaxis])


def _get_rotation_matrix(vel):
    """ Generates a rotation matrix which can be used to determine the
    corrected sensor offsets.

    In the 2d case this returns the following rotation matrix
    [cos[theta] -sin[theta]]
    [cos[theta]  sin[theta]]

    In the 2d case this will be a 3x3 matrix which rotates around the Z axis
    followed by a rotation about the new Y axis.

    Parameters
    ----------
    vel : np.arrary
        1xD vector denoting platform velocity in D dimensions

    Returns
    -------
    np.array
        DxD rotation matrix
    """
    if len(vel) == 3:
        return _rot3d(vel)
    elif len(vel) == 2:
        theta = _get_angle(vel, np.array([[1, 0]]))
        if vel[1] < 0:
            theta *= -1
        return np.array([[cos(theta), -sin(theta)],
                         [sin(theta), cos(theta)]])


def _get_angle(vec, axis):
    """ Returns the angle between a pair of vectors. Used to determine the
    angle of rotation required between relative rectangular cartesian
    coordinate frame of reference and platform inertial frame of reference.

    Parameters
    ----------
    vec : np.array
        1xD array denoting platform velocity
    axis : np.array
        Dx1 array denoting sensor offset relative to platform

    Returns
    -------
    Angle : float
        Angle, in radians, between the two vectors
    """
    vel_norm = vec / np.linalg.norm(vec)
    axis_norm = axis / np.linalg.norm(axis)

    return np.arccos(np.clip(np.dot(axis_norm, vel_norm), -1.0, 1.0))


def _rot3d(vec):
    """
    This approach determines the platforms attitude based upon its velocity
    component. It does not take into account potential platform roll, nor
    are the components calculated to account for physical artifacts such as
    platform trim (e.g. aircraft yaw whilst flying forwards).

    The process determines the yaw (x-y) and pitch (z to x-y plane) angles.
    The rotation matrix for a rotation by yaw around the Z-axis is then
    calculated, the rotated Y axis is then determined and used to calculate the
    rotation matrix which takes into account the platform pitch

    Parameters
    ----------
    vec: np.array
        platform velocity

    Returns
    -------
    np.array
        3x3 rotation matrix
    """
    # TODO handle platform roll
    yaw = np.arctan2(vec[[1]], vec[[0]])
    pitch = np.arctan2(vec[[2]],
                       np.sqrt(vec[[0]] ** 2 + vec[[1]] ** 2)) * -1
    rot_z = _rot_z(yaw)
    # Modify to correct for new y axis
    y_axis = np.array([0, 1, 0])
    rot_y = expm(np.cross(np.eye(3), np.dot(rot_z, y_axis) * pitch))

    return np.dot(rot_y, rot_z)


def _rot_z(theta):
    """ Returns a rotation matrix which will rotate a vector around the Z axis
    in a counter clockwise direction by theta radians.

    Parameters
    ----------
    theta : float
        Required rotation angle in radians

    Returns
    -------
    np.array
        3x3 rotation matrix
    """
    return np.array([[cos(theta), -sin(theta), 0],
                     [sin(theta), cos(theta), 0],
                     [0, 0, 1]])



# -*- coding: utf-8 -*-
from abc import abstractmethod

from ..base import Base


class MetricGenerator(Base):
    """Metric Generator base class

    Generates :class:`~.Metric` objects used to asses the performance of a
    tracker using data held in a :class:`~.MetricManager` object
    """

    @abstractmethod
    def compute_metric(self, manager, **kwargs):
        """Compute metric

        Parameters
        ----------
        manager : MetricManager
            containing the data to be used to create the metric(s)

        Returns
        -------
        : list of :class:`~.Metric` objects
            Generated metrics
        """
        raise NotImplementedError


class MetricManager(Base):
    """Metric Manager base class

    Holds the data and manages the production of :class:`~.Metric` objects
    through a :class:`~.MetricGenerator`
    """


class PlotGenerator(MetricGenerator):
    """PlotGenerator base class

    PlotGenerators generate metrics that are visualisations. Return
    :class:`~.PlottingMetric` objects.
    """



import datetime

import numpy as np
import pytest

from ...models.measurement.linear import LinearGaussian
from ...updater.kalman import KalmanUpdater
from ...types.detection import Detection
from ...types.state import GaussianState, ParticleState
from ..simple import (
    SinglePointInitiator, LinearMeasurementInitiator, GaussianParticleInitiator
)


def test_spi():
    """Test SinglePointInitiator"""

    # Prior state information
    prior_state = GaussianState(
        np.array([[0], [0]]),
        np.array([[100, 0], [0, 1]]))

    # Define a measurement model
    measurement_model = LinearGaussian(2, [0], np.array([[1]]))

    # Define the Initiator
    initiator = SinglePointInitiator(
        prior_state,
        measurement_model)

    # Define 2 detections from which tracks are to be initiated
    timestamp = datetime.datetime.now()
    detections = [Detection(np.array([[4.5]]), timestamp),
                  Detection(np.array([[-4.5]]), timestamp)]

    # Run the initiator based on the available detections
    tracks = initiator.initiate(detections)

    # Ensure same number of tracks are initiated as number of measurements
    # (i.e. 2)
    assert(len(tracks) == 2)

    # Ensure that tracks are initiated correctly
    evaluated_tracks = [False, False]
    for detection in detections:

        post_state_vec, post_state_covar, _, _, _, _ =\
            KalmanUpdater.update_lowlevel(prior_state.state_vector,
                                          prior_state.covar,
                                          measurement_model.matrix(),
                                          measurement_model.covar(),
                                          detection.state_vector)

        eval_track_state = GaussianState(
            post_state_vec,
            post_state_covar,
            timestamp=detection.timestamp)

        # Compare against both tracks
        for track_idx, track in enumerate(tracks):

            if(np.array_equal(eval_track_state.mean, track.mean)
               and np.array_equal(eval_track_state.covar, track.covar)):

                evaluated_tracks[track_idx] = True

    # Ensure both tracks have been evaluated
    assert(all(evaluated_tracks))

    assert set(detections) == set(track.state.hypothesis.measurement
                                  for track in tracks)


def test_linear_measurement():
    measurement_model = LinearGaussian(2, [0], np.array([[50]]))
    measurement_initiator = LinearMeasurementInitiator(
        GaussianState(np.array([[0], [0]]), np.diag([100, 10])),
        measurement_model
    )

    timestamp = datetime.datetime.now()
    detections = [Detection(np.array([[5]]), timestamp),
                  Detection(np.array([[-5]]), timestamp)]

    tracks = measurement_initiator.initiate(detections)

    for track in tracks:
        if track.state_vector[0, 0] > 0:
            assert np.array_equal(track.state_vector, np.array([[5], [0]]))
            assert np.array_equal(
                measurement_model.matrix()@track.state_vector,
                detections[0].state_vector)
            assert track.state.hypothesis.measurement is detections[0]
        else:
            assert np.array_equal(track.state_vector, np.array([[-5], [0]]))
            assert np.array_equal(
                measurement_model.matrix()@track.state_vector,
                detections[1].state_vector)
            assert track.state.hypothesis.measurement is detections[1]

        assert track.timestamp == timestamp

        assert np.array_equal(track.covar, np.diag([50, 10]))


def test_linear_measurement_non_direct():

    class _LinearMeasurementModel:
        ndim_state = 2
        ndmim_meas = 2

        @staticmethod
        def matrix():
            return np.array([[0, 1], [2, 0]])

        @staticmethod
        def covar():
            return np.diag([10, 50])

    measurement_model = _LinearMeasurementModel()
    measurement_initiator = LinearMeasurementInitiator(
        GaussianState(np.array([[0], [0]]), np.diag([100, 10])),
        measurement_model
    )

    timestamp = datetime.datetime.now()
    detections = [Detection(np.array([[5], [2]]), timestamp),
                  Detection(np.array([[-5], [8]]), timestamp)]

    tracks = measurement_initiator.initiate(detections)

    for track in tracks:
        if track.state_vector[1, 0] > 0:
            assert np.array_equal(track.state_vector, np.array([[1], [5]]))
            assert np.array_equal(
                measurement_model.matrix()@track.state_vector,
                detections[0].state_vector)
            assert track.state.hypothesis.measurement is detections[0]
        else:
            assert np.array_equal(track.state_vector, np.array([[4], [-5]]))
            assert np.array_equal(
                measurement_model.matrix()@track.state_vector,
                detections[1].state_vector)
            assert track.state.hypothesis.measurement is detections[1]

        assert track.timestamp == timestamp

        assert np.array_equal(track.covar, np.diag([25, 10]))


def test_linear_measurement_extra_state_dim():

    class _LinearMeasurementModel:
        ndim_state = 3
        ndmim_meas = 2

        @staticmethod
        def matrix():
            return np.array([[1, 0, 0], [0, 0, 1]])

        @staticmethod
        def covar():
            return np.diag([10, 50])

    measurement_model = _LinearMeasurementModel()
    measurement_initiator = LinearMeasurementInitiator(
        GaussianState(np.array([[0], [0], [0]]), np.diag([100, 10, 500])),
        measurement_model
    )

    timestamp = datetime.datetime.now()
    detections = [Detection(np.array([[5], [2]]), timestamp),
                  Detection(np.array([[-5], [8]]), timestamp)]

    tracks = measurement_initiator.initiate(detections)

    for track in tracks:
        if track.state_vector[0, 0] > 0:
            assert np.array_equal(
                track.state_vector,
                np.array([[5], [0], [2]]))
            assert np.array_equal(
                measurement_model.matrix()@track.state_vector,
                detections[0].state_vector)
            assert track.state.hypothesis.measurement is detections[0]
        else:
            assert np.array_equal(
                track.state_vector,
                np.array([[-5], [0], [8]]))
            assert np.array_equal(
                measurement_model.matrix()@track.state_vector,
                detections[1].state_vector)
            assert track.state.hypothesis.measurement is detections[1]

        assert track.timestamp == timestamp

        assert np.array_equal(track.covar, np.diag([10, 10, 50]))


@pytest.mark.parametrize("gaussian_initiator", [
    SinglePointInitiator(
        GaussianState(np.array([[0]]), np.array([[100]])),
        LinearGaussian(1, [0], np.array([[1]]))
    ),
    LinearMeasurementInitiator(
        GaussianState(np.array([[0]]), np.array([[100]])),
        LinearGaussian(1, [0], np.array([[1]]))
    ),
], ids=['SinglePoint', 'LinearMeasurement'])
def test_gaussian_particle(gaussian_initiator):
    particle_initiator = GaussianParticleInitiator(gaussian_initiator)

    timestamp = datetime.datetime.now()
    detections = [Detection(np.array([[5]]), timestamp),
                  Detection(np.array([[-5]]), timestamp)]

    tracks = particle_initiator.initiate(detections)

    for track in tracks:
        assert isinstance(track.state, ParticleState)
        if track.state_vector > 0:
            assert np.allclose(track.state_vector, np.array([[5]]), atol=0.4)
            assert track.state.hypothesis.measurement is detections[0]
        else:
            assert np.allclose(track.state_vector, np.array([[-5]]), atol=0.4)
            assert track.state.hypothesis.measurement is detections[1]
        assert track.timestamp == timestamp

        assert np.allclose(track.covar, np.array([[1]]), atol=0.4)



# -*- coding: utf-8 -*-
import copy
import numpy as np

from .base import Sensor
from ..base import Property
from ..models.measurement.nonlinear import RangeBearingGaussianToCartesian
from ..types.array import CovarianceMatrix
from ..types.detection import Detection
from ..types.state import State, StateVector


class RadarRangeBearing(Sensor):
    """A simple radar sensor that generates measurements of targets, using a
    :class:`~.RangeBearingGaussianToCartesian` model, relative to its position.

    Note
    ----
    The current implementation of this class assumes a 3D Cartesian plane.

    """

    position = Property(StateVector,
                        doc="The radar position on a 3D Cartesian plane,\
                             expressed as a 3x1 array of Cartesian coordinates\
                             in the order :math:`x,y,z`")
    orientation = Property(
        StateVector,
        doc="A 3x1 array of angles (rad), specifying the radar orientation in \
            terms of the counter-clockwise rotation around each Cartesian \
            axis in the order :math:`x,y,z`. The rotation angles are positive \
            if the rotation is in the counter-clockwise direction when viewed \
            by an observer looking along the respective rotation axis, \
            towards the origin")
    ndim_state = Property(
        int,
        doc="Number of state dimensions. This is utilised by (and follows in\
            format) the underlying :class:`~.RangeBearingGaussianToCartesian`\
            model")
    mapping = Property(
        [np.array], doc="Mapping between the targets state space and the\
                        sensors measurement capability")
    noise_covar = Property(CovarianceMatrix,
                           doc="The sensor noise covariance matrix. This is \
                                utilised by (and follow in format) the \
                                underlying \
                                :class:`~.RangeBearingGaussianToCartesian`\
                                model")

    def __init__(self, position, orientation, ndim_state, mapping, noise_covar,
                 *args, **kwargs):
        measurement_model = RangeBearingGaussianToCartesian(
            ndim_state=ndim_state,
            mapping=mapping,
            noise_covar=noise_covar,
            translation_offset=position,
            rotation_offset=orientation)

        super().__init__(position, orientation, ndim_state, mapping,
                         noise_covar, *args, measurement_model, **kwargs)

    def set_position(self, position):
        self.position = position
        self.measurement_model.translation_offset = position

    def set_orientation(self, orientation):
        self.orientation = orientation
        self.measurement_model.rotation_offset = orientation

    def gen_measurement(self, ground_truth, noise=None, **kwargs):
        """Generate a measurement for a given state

        Parameters
        ----------
        ground_truth : :class:`~.State`
            A ground-truth state

        Returns
        -------
        :class:`~.Detection`
            A measurement generated from the given state. The timestamp of the\
            measurement is set equal to that of the provided state.
        """

        measurement_vector = self.measurement_model.function(
            ground_truth.state_vector, noise=noise, **kwargs)

        model_copy = copy.copy(self.measurement_model)

        return Detection(measurement_vector,
                         measurement_model=model_copy,
                         timestamp=ground_truth.timestamp)


class RadarRotatingRangeBearing(RadarRangeBearing):
    """A simple rotating radar, with set field-of-view (FOV) angle, range and\
     rotations per minute (RPM), that generates measurements of targets, using\
     a :class:`~.RangeBearingGaussianToCartesian` model, relative to its\
     position.

    Note
    ----
    * The current implementation of this class assumes a 3D Cartesian plane.

    """

    dwell_center = Property(
        State, doc="A state object, whose `state_vector`\
        property describes the rotation angle of the center of the sensor's\
        current FOV (i.e. the dwell center) relative to the positive x-axis\
        of the sensor frame/orientation. The angle is positive if the rotation\
        is in the counter-clockwise direction when viewed by an observer\
        looking down the z-axis of the sensor frame, towards the origin.\
        Angle units are in radians"
    )
    rpm = Property(
        float, doc="The number of antenna rotations per minute (RPM)")
    max_range = Property(
        float, doc="The maximum detection range of the radar (in meters)")
    fov_angle = Property(
        float, doc="The radar field of view (FOV) angle (in radians).")

    def __init__(self, position, orientation, ndim_state, mapping, noise_covar,
                 dwell_center, rpm, max_range, fov_angle, *args, **kwargs):

        super().__init__(position, orientation, ndim_state, mapping,
                         noise_covar, dwell_center, rpm, max_range,
                         fov_angle, *args, **kwargs)

    def gen_measurement(self, ground_truth, noise=None, **kwargs):
        """Generate a measurement for a given state

        Parameters
        ----------
        ground_truth : :class:`~.State`
            A ground-truth state

        Returns
        -------
        :class:`~.Detection` or ``None``
            A measurement generated from the given state, if the state falls\
            in the sensor's field of view, or ``None``, otherwise. The\
            timestamp of the measurement is set equal to that of the provided\
            state.
        """

        # Read timestamp from ground truth
        timestamp = ground_truth.timestamp

        # Rotate the radar antenna and compute new heading
        self.rotate(timestamp)
        antenna_heading = self.orientation[2, 0] + \
            self.dwell_center.state_vector[0, 0]

        # Set rotation offset of underlying measurement model
        rot_offset =\
            StateVector(
                [[self.orientation[0, 0]],
                 [self.orientation[1, 0]],
                 [antenna_heading]])
        self.measurement_model.rotation_offset = rot_offset

        # Transform state to measurement space and generate
        # random noise
        measurement_vector = self.measurement_model.function(
            ground_truth.state_vector, noise=0, **kwargs)
        if(noise is None):
            measurement_noise = self.measurement_model.rvs()
        else:
            measurement_noise = noise

        # Check if state falls within sensor's FOV
        fov_min = -self.fov_angle/2
        fov_max = +self.fov_angle/2
        bearing_t = measurement_vector[0, 0]
        range_t = measurement_vector[1, 0]

        # Return None if state not in FOV
        if(bearing_t > fov_max or bearing_t < fov_min
           or range_t > self.max_range):
            return None

        # Else return measurement
        model_copy = copy.copy(self.measurement_model)
        measurement_vector += measurement_noise  # Add noise
        return Detection(measurement_vector,
                         measurement_model=model_copy,
                         timestamp=timestamp)

    def rotate(self, timestamp):
        """Rotate the sensor's antenna

        This method computes and updates the sensor's `dwell_center` property.

        Parameters
        ----------
        timestamp: :class:`datetime.datetime`
            A timestamp signifying when the rotation completes
        """

        # Compute duration since last rotation
        duration = timestamp - self.dwell_center.timestamp

        # Update dwell center
        rps = self.rpm/60  # rotations per sec
        self.dwell_center = State(
            StateVector([[self.dwell_center.state_vector[0, 0]
                          + duration.total_seconds()*rps*2*np.pi]]),
            timestamp
        )



# -*- coding: utf-8 -*-
import datetime

from ..time import UpdateTimeStepsDeleter, UpdateTimeDeleter
from ...types.detection import Detection
from ...types.hypothesis import SingleHypothesis
from ...types.prediction import StatePrediction
from ...types.track import Track
from ...types.update import StateUpdate


def test_update_time_steps_deleter():
    deleter = UpdateTimeStepsDeleter(3)

    track = Track([
        StateUpdate(
            [[0]],
            SingleHypothesis(None, Detection([[0]])),
            timestamp=datetime.datetime(2018, 1, 1, 14)),
        StatePrediction(
            [[0]], timestamp=datetime.datetime(2018, 1, 1, 14, 10)),
        StatePrediction(
            [[0]], timestamp=datetime.datetime(2018, 1, 1, 14, 20)),
    ])

    # Last update within time steps
    tracks2delete = deleter.delete_tracks({track})
    assert not tracks2delete

    # Add new state (taking outside time)
    track.append(StatePrediction(
        [[0]], timestamp=datetime.datetime(2018, 1, 1, 14, 30)))

    # Last update outside time step
    tracks2delete = deleter.delete_tracks({track})
    assert tracks2delete

    # Add new update without measurement
    track.append(StateUpdate(
        [[0]],
        SingleHypothesis(None, None),
        timestamp=datetime.datetime(2018, 1, 1, 14, 30)))
    tracks2delete = deleter.delete_tracks({track})
    assert tracks2delete


def test_update_time_deleter():
    deleter = UpdateTimeDeleter(datetime.timedelta(minutes=20))

    track = Track([
        StateUpdate(
            [[0]],
            SingleHypothesis(None, Detection([[0]])),
            timestamp=datetime.datetime(2018, 1, 1, 14)),
        StatePrediction(
            [[0]], timestamp=datetime.datetime(2018, 1, 1, 14, 10)),
        StatePrediction(
            [[0]], timestamp=datetime.datetime(2018, 1, 1, 14, 20)),
    ])

    # Last update within time
    tracks2delete = deleter.delete_tracks({track})
    assert not tracks2delete

    # Last update within custom time
    tracks2delete = deleter.delete_tracks(
        {track}, timestamp=datetime.datetime(2018, 1, 1, 14, 15))
    assert not tracks2delete

    # Last update outside custom time
    tracks2delete = deleter.delete_tracks(
        {track}, timestamp=datetime.datetime(2018, 1, 1, 14, 25))
    assert tracks2delete

    # Add new state (taking outside time)
    track.append(StatePrediction(
        [[0]], timestamp=datetime.datetime(2018, 1, 1, 14, 30)))

    # Last update outside time
    tracks2delete = deleter.delete_tracks({track})
    assert tracks2delete

    # Add new update without measurement
    track.append(StateUpdate(
        [[0]],
        SingleHypothesis(None, None),
        timestamp=datetime.datetime(2018, 1, 1, 14, 30)))
    tracks2delete = deleter.delete_tracks({track})
    assert tracks2delete

    # Last update within custom time
    tracks2delete = deleter.delete_tracks(
        {track}, timestamp=datetime.datetime(2018, 1, 1, 14, 15))
    assert not tracks2delete



import datetime
from operator import attrgetter

import numpy as np

from ..distance import DistanceHypothesiser
from ..filtered import FilteredDetectionsHypothesiser
from ...types.detection import Detection
from ...types.hypothesis import SingleHypothesis
from ...types.track import Track
from ...types.update import GaussianStateUpdate
from ... import measures as measures

measure = measures.Mahalanobis()


def test_filtereddetections(predictor, updater):
    # CASE 1
    # there is one track with associated metadata,
    # two detections where one has matching metadata and one does not

    timestamp = datetime.datetime.now()

    hypothesiser = DistanceHypothesiser(predictor, updater,
                                        measure=measure, missed_distance=0.2)
    hypothesiser_wrapper = FilteredDetectionsHypothesiser(
        hypothesiser, "MMSI", match_missing=True)

    track = Track([GaussianStateUpdate(
                    np.array([[0]]),
                    np.array([[1]]),
                    SingleHypothesis(
                        None,
                        Detection(np.array([[0]]), metadata={"MMSI": 12345})),
                    timestamp=timestamp)])
    detection1 = Detection(np.array([[2]]), metadata={"MMSI": 12345})
    detection2 = Detection(np.array([[3]]), metadata={"MMSI": 99999})
    detections = {detection1, detection2}

    hypotheses = hypothesiser_wrapper.hypothesise(track, detections, timestamp)

    # There are 2 hypotheses - Detection 1,  Missed Dectection
    # - Detection 2 has different metadata, so no hypothesis
    assert len(hypotheses) == 2
    assert all(not hypothesis.measurement or
               hypothesis.measurement.metadata['MMSI'] == 12345
               for hypothesis in hypotheses)

    # There is a missed detection hypothesis
    assert any(not hypothesis.measurement for hypothesis in hypotheses)

    # Each hypothesis has a distance attribute
    assert all(hypothesis.distance >= 0 for hypothesis in hypotheses)

    # The hypotheses are sorted correctly
    assert min(hypotheses, key=attrgetter('distance')) is hypotheses[0]


def test_filtereddetections_empty_detections(predictor, updater):
    # CASE 3
    # 'detections' is empty
    timestamp = datetime.datetime.now()
    hypothesiser = DistanceHypothesiser(predictor, updater,
                                        measure=measure, missed_distance=0.2)
    hypothesiser_wrapper = FilteredDetectionsHypothesiser(
        hypothesiser, "MMSI", match_missing=False)

    track = Track([GaussianStateUpdate(
        np.array([[0]]),
        np.array([[1]]),
        SingleHypothesis(
            None,
            Detection(np.array([[0]]), metadata={"MMSI": 12345})),
        timestamp=timestamp)])
    detections = {}

    hypotheses = hypothesiser_wrapper.hypothesise(track, detections, timestamp)

    # one hypothesis - Missed Detection
    assert len(hypotheses) == 1

    # There is a missed detection hypothesis
    assert any(not hypothesis.measurement for hypothesis in hypotheses)


def test_filtereddetections_no_track_metadata(predictor, updater):
    # CASE 2
    # there is one track with NO associated metadata, two detections
    # with metadata, hypothesiser.match_missing is True (default), so
    # detections with any metadata can be associated with the track

    timestamp = datetime.datetime.now()
    hypothesiser = DistanceHypothesiser(predictor, updater,
                                        measure=measure, missed_distance=0.2)
    hypothesiser_wrapper = FilteredDetectionsHypothesiser(
        hypothesiser, "MMSI", match_missing=True)

    track = Track([GaussianStateUpdate(
        np.array([[0]]),
        np.array([[1]]),
        SingleHypothesis(
            None,
            Detection(np.array([[0]]), metadata={})),
        timestamp=timestamp)])
    detection1 = Detection(np.array([[2]]), metadata={"MMSI": 12345})
    detection2 = Detection(np.array([[3]]), metadata={"MMSI": 99999})
    detections = {detection1, detection2}

    hypotheses = hypothesiser_wrapper.hypothesise(track, detections, timestamp)

    # There are 3 hypotheses - Detection 1, Detection 2, Missed Detection
    assert len(hypotheses) == 3
    assert all(not hypothesis.measurement or
               hypothesis.measurement.metadata['MMSI'] == 12345 or
               hypothesis.measurement.metadata['MMSI'] == 99999
               for hypothesis in hypotheses)

    # There is a missed detection hypothesis
    assert any(not hypothesis.measurement for hypothesis in hypotheses)

    # Each hypothesis has a distance attribute
    assert all(hypothesis.distance >= 0 for hypothesis in hypotheses)

    # The hypotheses are sorted correctly
    assert min(hypotheses, key=attrgetter('distance')) is hypotheses[0]


def test_filtereddetections_no_matching_metadata(predictor, updater):
    # CASE 4
    # there is one track with associated metadata,
    # two detections where neither has matching metadata

    timestamp = datetime.datetime.now()
    hypothesiser = DistanceHypothesiser(predictor, updater,
                                        measure=measure, missed_distance=0.2)
    hypothesiser_wrapper = FilteredDetectionsHypothesiser(
        hypothesiser, "MMSI", match_missing=True)

    track = Track([GaussianStateUpdate(
                    np.array([[0]]),
                    np.array([[1]]),
                    SingleHypothesis(
                        None,
                        Detection(np.array([[0]]), metadata={"MMSI": 12345})),
                    timestamp=timestamp)])
    detection1 = Detection(np.array([[2]]), metadata={"MMSI": 45678})
    detection2 = Detection(np.array([[3]]), metadata={"MMSI": 99999})
    detections = {detection1, detection2}

    hypotheses = hypothesiser_wrapper.hypothesise(track, detections, timestamp)

    # one hypothesis - Missed Detection
    assert len(hypotheses) == 1

    # There is a missed detection hypothesis
    assert any(not hypothesis.measurement for hypothesis in hypotheses)



# -*- coding: utf-8 -*-
from ..base import Property
from .array import CovarianceMatrix
from .base import Type
from .state import State, GaussianState, ParticleState


class Prediction(Type):
    """ Prediction type

    This is the base prediction class. """


class MeasurementPrediction(Type):
    """ Prediction type

    This is the base measurement prediction class. """


class StatePrediction(State, Prediction):
    """ StatePrediction type

    Most simple state prediction type, which only has time and a state vector.
    """


class StateMeasurementPrediction(State, MeasurementPrediction):
    """ MeasurementPrediction type

    Most simple measurement prediction type, which only has time and a state
    vector.
    """


class GaussianStatePrediction(Prediction, GaussianState):
    """ GaussianStatePrediction type

    This is a simple Gaussian state prediction object, which, as the name
    suggests, is described by a Gaussian distribution.
    """


class GaussianMeasurementPrediction(MeasurementPrediction, GaussianState):
    """ GaussianMeasurementPrediction type

    This is a simple Gaussian measurement prediction object, which, as the name
    suggests, is described by a Gaussian distribution.
    """

    cross_covar = Property(CovarianceMatrix,
                           doc="The state-measurement cross covariance matrix",
                           default=None)

    def __init__(self, state_vector, covar, timestamp=None,
                 cross_covar=None, *args, **kwargs):
        if(cross_covar is not None
           and cross_covar.shape[1] != state_vector.shape[0]):
            raise ValueError("cross_covar should have the same number of \
                             columns as the number of rows in state_vector")
        super().__init__(state_vector, covar, timestamp,
                         cross_covar, *args, **kwargs)


class ParticleStatePrediction(Prediction, ParticleState):
    """ParticleStatePrediction type

    This is a simple Particle state prediction object.
    """


class ParticleMeasurementPrediction(MeasurementPrediction, ParticleState):
    """MeasurementStatePrediction type

    This is a simple Particle measurement prediction object.
    """



# -*- coding: utf-8 -*-
from abc import abstractmethod

from ..base import Base, Property
from ..models.transition import TransitionModel


class Smoother(Base):
    """Smoother Base Class."""

    transition_model = Property(
        TransitionModel, default=None, doc="Transition Model.")

    @abstractmethod
    def smooth(self, *args, **kwargs):
        raise NotImplementedError

    @abstractmethod
    def track_smooth(self, *args, **kwargs):
        raise NotImplementedError



import datetime
import numpy as np
import matplotlib.figure

from ..plotter import TwoDPlotter
from ...types.track import Track
from ...types.groundtruth import GroundTruthPath
from ...types.state import State
from ...types.detection import Detection


def test_twodplotter():
    plotter = TwoDPlotter(track_indices=[0, 1],
                          detection_indices=[0, 1],
                          gtruth_indices=[0, 1])
    timestamp1 = datetime.datetime.now()
    timestamp2 = timestamp1 + datetime.timedelta(seconds=10)

    tracks = {Track(states=[State(np.array([[1], [2]]),
                                  timestamp=timestamp1 + datetime.timedelta(
                                      seconds=i)) for i in range(11)])}
    truths = {GroundTruthPath(states=[State(np.array([[1], [2]]),
                                            timestamp=timestamp1 +
                                            datetime.timedelta(seconds=i))
                                      for i in range(11)])}
    dets = {Detection(np.array([[1], [2]]),
                      timestamp=timestamp1+datetime.timedelta(seconds=i))
            for i in range(11)}

    metric = plotter.plot_tracks_truth_detections(tracks, truths, dets)

    assert metric.title == "Track plot"
    assert metric.generator == plotter
    assert type(metric.value) == matplotlib.figure.Figure
    assert metric.time_range.start_timestamp == timestamp1
    assert metric.time_range.end_timestamp == timestamp2



# -*- coding: utf-8 -*-
import datetime

from .base import Type
from .time import TimeRange
from ..base import Property


class Metric(Type):
    """Metric type"""

    title = Property(str, doc='Name of the metric')
    value = Property(any, doc='Value of the metric')
    generator = Property(any, doc='Generator used to create the metric')


class PlottingMetric(Metric):
    """Metric which is to be visualised as plot, value should be a pyplot
     figure"""


class SingleTimeMetric(Metric):
    """Metric for a specific timestamp"""

    timestamp = Property(datetime.datetime, default=None,
                         doc="Timestamp of the state. Default None.")


class TimeRangeMetric(Metric):
    """ Metric for a range of times (e.g. for example an entire run)"""

    time_range = Property(TimeRange, default=None,
                          doc="Time range over which metric assessment will be"
                              "conducted over. Default is None")


class TimeRangePlottingMetric(TimeRangeMetric, PlottingMetric):
    """Plotting metric covering a period of time"""


class SingleTimePlottingMetric(SingleTimeMetric, PlottingMetric):
    """Plotting metric covering a specific timestamp"""



# -*- coding: utf-8 -*-

from ..base import Property
from .array import StateVector
from .base import Type


class Particle(Type):
    """
    Particle type

    A particle type which contains a state and weight
    """
    state_vector = Property(StateVector, doc="State vector")
    weight = Property(float, doc='Weight of particle')
    parent = Property(None, default=None, doc='Parent particle')

    def __init__(self, state_vector, weight, parent=None, *args, **kwargs):
        if parent:
            parent.parent = None
        super().__init__(state_vector, weight, parent, *args, **kwargs)
Particle.parent.cls = Particle  # noqa:E305



# -*- coding: utf-8 -*-

import scipy as sp
from scipy.stats import multivariate_normal

from ...base import Property
from ...functions import cart2pol, cart2sphere, cart2angles, rotx, roty, rotz
from ...types.array import StateVector, CovarianceMatrix
from ...types.angle import Bearing, Elevation
from ..base import NonLinearModel, GaussianModel
from .base import MeasurementModel


class NonLinearGaussianMeasurement(MeasurementModel,
                                   NonLinearModel,
                                   GaussianModel):
    r"""This class combines the MeasurementModel, NonLinearModel and \
    GaussianModel classes. It is not meant to be instantiated directly \
    but subclasses should be derived from this class.
    """
    noise_covar = Property(CovarianceMatrix, doc="Noise covariance")
    rotation_offset = Property(
        StateVector, default=StateVector(sp.array([[0], [0], [0]])),
        doc="A 3x1 array of angles (rad), specifying the clockwise rotation\
            around each Cartesian axis in the order :math:`x,y,z`.\
            The rotation angles are positive if the rotation is in the \
            counter-clockwise direction when viewed by an observer looking\
            along the respective rotation axis, towards the origin.")

    def covar(self, **kwargs):
        """Returns the measurement model noise covariance matrix.

        Returns
        -------
        :class:`~.CovarianceMatrix` of shape\
        (:py:attr:`~ndim_meas`, :py:attr:`~ndim_meas`)
            The measurement noise covariance.
        """

        return self.noise_covar

    def pdf(self, meas_vec, state_vec, **kwargs):
        r""" Measurement pdf/likelihood evaluation function

        Evaluates the pdf/likelihood of the (set of) measurement vector(s)
        ``meas_vec``, given the (set of) state vector(s) ``state_vec``.

        In mathematical terms, this can be written as:

        .. math::

            p(\vec{y}_t | \vec{x}_t) = \mathcal{N}(\vec{y}_t; \vec{x}_t, R)

        Parameters
        ----------
        meas_vec : :class:`~.StateVector`
            A measurement
        state_vec : :class:`~.StateVector`
            A state

        Returns
        -------
        :class:`float`
            The likelihood of ``meas``, given ``state``
        """

        likelihood = multivariate_normal.pdf(
            meas_vec.T,
            mean=(self.function(state_vec, 0)).ravel(),
            cov=self.covar()
        )
        return likelihood

    def rvs(self, num_samples=1, **kwargs):
        r""" Model noise/sample generation function

        Generates noise samples from the measurement model.

        In mathematical terms, this can be written as:

        .. math::

            \vec{v}_t \sim \mathcal{N}(0,R)

        Parameters
        ----------
        num_samples: scalar, optional
            The number of samples to be generated (the default is 1)

        Returns
        -------
        2-D array of shape (:py:attr:`~ndim_meas`, ``num_samples``)
            A set of Np samples, generated from the model's noise
            distribution.
        """

        noise = multivariate_normal.rvs(
            sp.zeros(self.ndim_meas), self.covar(), num_samples)

        if num_samples == 1:
            return noise.reshape((-1, 1))
        else:
            return noise.T

    @property
    def _rotation_matrix(self):
        """_rotation_matrix getter method

        Calculates and returns the (3D) axis rotation matrix.

        Returns
        -------
        :class:`numpy.ndarray` of shape (3, 3)
            The model (3D) rotation matrix.
        """

        theta_x = -self.rotation_offset[0, 0]
        theta_y = -self.rotation_offset[1, 0]
        theta_z = -self.rotation_offset[2, 0]

        return rotz(theta_z)@roty(theta_y)@rotx(theta_x)


class RangeBearingElevationGaussianToCartesian(NonLinearGaussianMeasurement):
    r"""This is a class implementation of a time-invariant measurement model, \
    where measurements are assumed to be received in the form of bearing \
    (:math:`\phi`), elevation (:math:`\theta`) and range (:math:`r`), with \
    Gaussian noise in each dimension.

    The model is described by the following equations:

    .. math::

      \vec{y}_t = h(\vec{x}_t, \vec{v}_t)

    where:

    * :math:`\vec{y}_t` is a measurement vector of the form:

    .. math::

      \vec{y}_t = \begin{bmatrix}
                \theta \\
                \phi \\
                r
            \end{bmatrix}

    * :math:`h` is a non-linear model function of the form:

    .. math::

      h(\vec{x}_t,\vec{v}_t) = \begin{bmatrix}
                asin(\mathcal{z}/\sqrt{\mathcal{x}^2 + \mathcal{y}^2 +\mathcal{z}^2}) \\
                atan2(\mathcal{y},\mathcal{x}) \\
                \sqrt{\mathcal{x}^2 + \mathcal{y}^2 + \mathcal{z}^2}
                \end{bmatrix} + \vec{v}_t

    * :math:`\vec{v}_t` is Gaussian distributed with covariance :math:`R`, i.e.:

    .. math::

      \vec{v}_t \sim \mathcal{N}(0,R)

    .. math::

      R = \begin{bmatrix}
            \sigma_{\theta}^2 & 0 & 0 \\
            0 & \sigma_{\phi}^2 & 0 \\
            0 & 0 & \sigma_{r}^2
            \end{bmatrix}

    The :py:attr:`mapping` property of the model is a 3 element vector, \
    whose first (i.e. :py:attr:`mapping[0]`), second (i.e. \
    :py:attr:`mapping[1]`) and third (i.e. :py:attr:`mapping[2`) elements \
    contain the state index of the :math:`x`, :math:`y` and :math:`z`  \
    coordinates, respectively.

    Note
    ----
    The current implementation of this class assumes a 3D Cartesian plane.

    """  # noqa:E501

    translation_offset = Property(
        StateVector, default=StateVector(sp.array([[0], [0], [0]])),
        doc="A 3x1 array specifying the Cartesian origin offset in terms of :math:`x,y,z`\
            coordinates.")

    @property
    def ndim_meas(self):
        """ndim_meas getter method

        Returns
        -------
        :class:`int`
            The number of measurement dimensions
        """

        return 3

    def function(self, state_vector, noise=None, **kwargs):
        r"""Model function :math:`h(\vec{x}_t,\vec{v}_t)`

        Parameters
        ----------
        state_vector: :class:`~.StateVector`
            An input state vector
        noise: :class:`numpy.ndarray`
            An externally generated random process noise sample (the default in
            `None`, in which case process noise will be generated internally)

        Returns
        -------
        :class:`numpy.ndarray` of shape (:py:attr:`~ndim_state`, 1)
            The model function evaluated given the provided time interval.
        """

        if noise is None:
            noise = self.rvs()

        # Account for origin offset
        xyz = state_vector[self.mapping] - self.translation_offset

        # Rotate coordinates
        xyz_rot = self._rotation_matrix @ xyz

        # Convert to Spherical
        rho, phi, theta = cart2sphere(*xyz_rot[:, 0])

        return sp.array([[Elevation(theta)],
                         [Bearing(phi)],
                         [rho]]) + noise

    def rvs(self, num_samples=1, **kwargs):
        out = super().rvs(num_samples, **kwargs)
        out = sp.array([[Elevation(0.)], [Bearing(0.)], [0.]]) + out
        return out


class RangeBearingGaussianToCartesian(
        NonLinearGaussianMeasurement):
    r"""This is a class implementation of a time-invariant measurement model, \
    where measurements are assumed to be received in the form of bearing \
    (:math:`\phi`) and range (:math:`r`), with Gaussian noise in each dimension.

    The model is described by the following equations:

    .. math::

      \vec{y}_t = h(\vec{x}_t, \vec{v}_t)

    where:

    * :math:`\vec{y}_t` is a measurement vector of the form:

    .. math::

      \vec{y}_t = \begin{bmatrix}
                \phi \\
                r
            \end{bmatrix}

    * :math:`h` is a non-linear model function of the form:

    .. math::

      h(\vec{x}_t,\vec{v}_t) = \begin{bmatrix}
                atan2(\mathcal{y},\mathcal{x}) \\
                \sqrt{\mathcal{x}^2 + \mathcal{y}^2}
                \end{bmatrix} + \vec{v}_t

    * :math:`\vec{v}_t` is Gaussian distributed with covariance :math:`R`, i.e.:

    .. math::

      \vec{v}_t \sim \mathcal{N}(0,R)

    .. math::

      R = \begin{bmatrix}
            \sigma_{\phi}^2 & 0 \\
            0 & \sigma_{r}^2
            \end{bmatrix}

    The :py:attr:`mapping` property of the model is a 2 element vector, \
    whose first (i.e. :py:attr:`mapping[0]`) and second (i.e. \
    :py:attr:`mapping[0]`) elements contain the state index of the \
    :math:`x` and :math:`y` coordinates, respectively.

    Note
    ----
    The current implementation of this class assumes a 2D Cartesian plane.

    """  # noqa:E501

    translation_offset = Property(
        StateVector, default=StateVector(sp.array([[0], [0]])),
        doc="A 2x1 array specifying the origin offset in terms of :math:`x,y`\
            coordinates.")

    @property
    def ndim_meas(self):
        """ndim_meas getter method

        Returns
        -------
        :class:`int`
            The number of measurement dimensions
        """

        return 2

    def function(self, state_vector, noise=None, **kwargs):
        r"""Model function :math:`h(\vec{x}_t,\vec{v}_t)`

        Parameters
        ----------
        state_vector: :class:`~.StateVector`
            An input state vector
        noise: :class:`numpy.ndarray`
            An externally generated random process noise sample (the default in
            `None`, in which case process noise will be generated internally)

        Returns
        -------
        :class:`numpy.ndarray` of shape (:py:attr:`~ndim_meas`, 1)
            The model function evaluated given the provided time interval.
        """

        if noise is None:
            noise = self.rvs()

        # Account for origin offset
        xyz = [[state_vector[self.mapping[0], 0]
                - self.translation_offset[0, 0]],
               [state_vector[self.mapping[1], 0]
                - self.translation_offset[1, 0]],
               [0]]

        # Rotate coordinates
        xyz_rot = self._rotation_matrix @ xyz

        # Covert to polar
        rho, phi = cart2pol(*xyz_rot[:2, 0])

        return sp.array([[Bearing(phi)], [rho]]) + noise

    def rvs(self, num_samples=1, **kwargs):
        out = super().rvs(num_samples, **kwargs)
        out = sp.array([[Bearing(0)], [0.]]) + out
        return out


class BearingElevationGaussianToCartesian(NonLinearGaussianMeasurement):
    r"""This is a class implementation of a time-invariant measurement model, \
    where measurements are assumed to be received in the form of bearing \
    (:math:`\phi`) and elevation (:math:`\theta`) and with \
    Gaussian noise in each dimension.

    The model is described by the following equations:

    .. math::

      \vec{y}_t = h(\vec{x}_t, \vec{v}_t)

    where:

    * :math:`\vec{y}_t` is a measurement vector of the form:

    .. math::

      \vec{y}_t = \begin{bmatrix}
                \theta \\
                \phi
            \end{bmatrix}

    * :math:`h` is a non-linear model function of the form:

    .. math::

      h(\vec{x}_t,\vec{v}_t) = \begin{bmatrix}
                asin(\mathcal{z}/\sqrt{\mathcal{x}^2 + \mathcal{y}^2 +\mathcal{z}^2}) \\
                atan2(\mathcal{y},\mathcal{x}) \\
                \end{bmatrix} + \vec{v}_t

    * :math:`\vec{v}_t` is Gaussian distributed with covariance :math:`R`, i.e.:

    .. math::

      \vec{v}_t \sim \mathcal{N}(0,R)

    .. math::

      R = \begin{bmatrix}
            \sigma_{\theta}^2 & 0 \\
            0 & \sigma_{\phi}^2\\
            \end{bmatrix}

    The :py:attr:`mapping` property of the model is a 3 element vector, \
    whose first (i.e. :py:attr:`mapping[0]`), second (i.e. \
    :py:attr:`mapping[1]`) and third (i.e. :py:attr:`mapping[2]`) elements  \
    contain the state index of the :math:`x`, :math:`y` and :math:`z`  \
    coordinates, respectively.

    Note
    ----
    The current implementation of this class assumes a 3D Cartesian plane.

    """  # noqa:E501

    translation_offset = Property(
        StateVector, default=StateVector(sp.array([[0], [0], [0]])),
        doc="A 3x1 array specifying the origin offset in terms of :math:`x,y,z`\
            coordinates.")

    @property
    def ndim_meas(self):
        """ndim_meas getter method

        Returns
        -------
        :class:`int`
            The number of measurement dimensions
        """

        return 2

    def function(self, state_vector, noise=None, **kwargs):
        r"""Model function :math:`h(\vec{x}_t,\vec{v}_t)`

        Parameters
        ----------
        state_vector: :class:`~.StateVector`
            An input state vector
        noise: :class:`numpy.ndarray`
            An externally generated random process noise sample (the default in
            `None`, in which case process noise will be generated internally)

        Returns
        -------
        :class:`numpy.ndarray` of shape (:py:attr:`~ndim_state`, 1)
            The model function evaluated given the provided time interval.
        """

        if noise is None:
            noise = self.rvs()

        # Account for origin offset
        xyz = state_vector[self.mapping] - self.translation_offset

        # Rotate coordinates
        xyz_rot = self._rotation_matrix @ xyz

        # Convert to Angles
        phi, theta = cart2angles(*xyz_rot[:, 0])

        return sp.array([[Elevation(theta)],
                         [Bearing(phi)]]) + noise

    def rvs(self, num_samples=1, **kwargs):
        out = super().rvs(num_samples, **kwargs)
        out = sp.array([[Elevation(0.)], [Bearing(0.)]]) + out
        return out
